# Copyright 2022 Planet Labs, PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
from datetime import datetime
import logging
import pytest

from planet import exceptions, io

LOGGER = logging.getLogger(__name__)


@pytest.fixture
def make_aiter():

    async def func(values):
        for v in values:
            yield v

    return func


@pytest.mark.anyio
async def test_collect_non_features(make_aiter):
    values = [{
        'key11': 'value11', 'key12': 'value12'
    }, {
        'key21': 'value21', 'key22': 'value22'
    }]

    values_aiter = make_aiter(values)
    res = await io.collect(values_aiter)
    assert res == values


@pytest.mark.anyio
async def test_collect_features(feature_geojson, make_aiter):
    feature2 = feature_geojson.copy()
    feature2['properties'] = {'foo': 'bar'}
    values = [feature_geojson, feature2]
    values_aiter = make_aiter(values)

    res = await io.collect(values_aiter)
    expected = {'type': 'FeatureCollection', 'features': values}
    assert res == expected


@pytest.mark.parametrize(
    "string, expected",
    [
        # RFC 3339 string from API response
        ("2021-02-03T01:40:07.359Z", datetime(2021, 2, 3, 1, 40, 7, 359000)),
        # date-only ISO 8601
        ("2021-01-01", datetime(2021, 1, 1)),
        # date-time ISO 8601
        ("2021-01-01", datetime(2021, 1, 1))
    ])
def test_str_to_datetime_success(string, expected):
    assert io.str_to_datetime(string) == expected


def test_str_to_datetime_failure():
    with pytest.raises(exceptions.PlanetError):
        io.str_to_datetime('notadate')



# Copyright 2021 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Functionality for reporting progress."""
import logging
from typing import Optional

from tqdm.asyncio import tqdm

LOGGER = logging.getLogger(__name__)


class ProgressBar:
    """Abstract base class for progress bar reporters."""

    def __init__(self, disable: bool = False):
        self.bar = None
        self.disable = disable

    def __str__(self):
        return str(self.bar)

    def __enter__(self):
        self.open_bar()
        return self

    def __exit__(self, *args):
        self.bar.close()

    def open_bar(self):
        """Initialize and start the progress bar."""
        return NotImplementedError


class StateBar(ProgressBar):
    """Bar reporter of order state.

    Example:
        ```python
        from planet import reporting

        with reporting.StateBar(state='creating') as bar:
            bar.update(state='created', order_id='oid')
            ...
        ```
    """

    def __init__(
        self,
        order_id: Optional[str] = None,
        state: Optional[str] = None,
        disable: bool = False,
    ):
        """Initialize the object.

        Parameters:
            order_id: Id of the order.
            state: State of the order.
        """

        self.state = state or ''
        self.order_id = order_id or ''
        super().__init__(disable=disable)

    def open_bar(self):
        """Initialize and start the progress bar."""
        self.bar = tqdm(
            bar_format="{elapsed} - {desc} - {postfix[0]}: {postfix[1]}",
            desc=self.desc,
            postfix=["state", self.state],
            disable=self.disable)

    @property
    def desc(self):
        return f'order {self.order_id}'

    def update_state(self, state: str):
        """Simple function to be used as a callback for state reporting"""
        self.update(state=state)

    def update(self,
               state: Optional[str] = None,
               order_id: Optional[str] = None):
        if state:
            self.state = state
            if self.bar is not None:
                try:
                    self.bar.postfix[1] = self.state
                except AttributeError:
                    # If the bar is disabled, attempting to access
                    # self.bar.postfix will result in an error. In this
                    # case, just skip it.
                    pass

        if order_id:
            self.order_id = order_id
            if self.bar is not None:
                self.bar.set_description_str(self.desc, refresh=False)

        if self.bar is not None:
            self.bar.refresh()


class AssetStatusBar(ProgressBar):
    """Bar reporter of asset status."""

    def __init__(
        self,
        item_type,
        item_id,
        asset_type,
        disable: bool = False,
    ):
        """Initialize the object.
        """
        self.item_type = item_type
        self.item_id = item_id
        self.asset_type = asset_type
        self.status = ''
        super().__init__(disable=disable)

    def open_bar(self):
        """Initialize and start the progress bar."""
        self.bar = tqdm(
            bar_format="{elapsed} - {desc} - {postfix[0]}: {postfix[1]}",
            desc=self.desc,
            postfix=["status", self.status],
            disable=self.disable)

    @property
    def desc(self):
        return f'{self.item_type} {self.item_id} {self.asset_type}'

    def update(self, status: str):
        self.status = status

        if self.bar is not None:
            try:
                self.bar.postfix[1] = self.status
            except AttributeError:
                # If the bar is disabled, attempting to access self.bar.postfix
                # will result in an error. In this case, just skip it.
                pass

        if self.bar is not None:
            self.bar.refresh()



# Copyright 2020 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Example of creating and downloading multiple orders.

This is an example of submitting two orders, waiting for them to complete, and
downloading them. The orders each clip a set of images to a specific area of
interest (AOI), so they cannot be combined into one order.

[Planet Explorer](https://www.planet.com/explorer/) was used to define
the AOIs and get the image ids.
"""
import asyncio
import os

import planet

DOWNLOAD_DIR = os.getenv('TEST_DOWNLOAD_DIR', '.')


def create_requests():
    # The Orders API will be asked to mask, or clip, results to
    # this area of interest.
    iowa_aoi = {
        "type": "Polygon",
        "coordinates": [[[-91.198465, 42.893071], [-91.121931, 42.893071],
                         [-91.121931, 42.946205], [-91.198465, 42.946205],
                         [-91.198465, 42.893071]]]
    }

    # In practice, you will use a Data API search to find items, but
    # for this example take them as given.
    iowa_items = ['20200925_161029_69_2223', '20200925_161027_48_2223']

    iowa_order = planet.order_request.build_request(
        name='iowa_order',
        products=[
            planet.order_request.product(item_ids=iowa_items,
                                         product_bundle='analytic_udm2',
                                         item_type='PSScene')
        ],
        tools=[planet.order_request.clip_tool(aoi=iowa_aoi)])

    oregon_aoi = {
        "type": "Polygon",
        "coordinates": [[[-117.558734, 45.229745], [-117.452447, 45.229745],
                         [-117.452447, 45.301865], [-117.558734, 45.301865],
                         [-117.558734, 45.229745]]]
    }

    oregon_items = ['20200909_182525_1014', '20200909_182524_1014']

    oregon_order = planet.order_request.build_request(
        name='oregon_order',
        products=[
            planet.order_request.product(item_ids=oregon_items,
                                         product_bundle='analytic_udm2',
                                         item_type='PSScene')
        ],
        tools=[planet.order_request.clip_tool(aoi=oregon_aoi)])

    return [iowa_order, oregon_order]


async def create_and_download(client, order_detail, directory):
    """Make an order, wait for completion, download files as a single task."""
    with planet.reporting.StateBar(state='creating') as reporter:
        order = await client.create_order(order_detail)
        reporter.update(state='created', order_id=order['id'])
        await client.wait(order['id'], callback=reporter.update_state)

    await client.download_order(order['id'], directory, progress_bar=True)


async def main():
    async with planet.Session() as sess:
        client = sess.client('orders')

        requests = create_requests()

        await asyncio.gather(*[
            create_and_download(client, request, DOWNLOAD_DIR)
            for request in requests
        ])


if __name__ == '__main__':
    asyncio.run(main())



# Copyright 2025 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

import logging
from typing import Any, AsyncIterator, Awaitable, Optional, Union, TypeVar

from planet.http import Session
from planet.models import Feature, GeoInterface, Paged
from planet.constants import PLANET_BASE_URL

T = TypeVar("T")

BASE_URL = f'{PLANET_BASE_URL}/features/v1/ogc/my/'

LOGGER = logging.getLogger()


class FeaturesClient:
    """Asyncronous Features API client

    For more information about the Features API, see the documentation at
    https://developers.planet.com/docs/apis/features/

    Example:
        ```python
        >>> import asyncio
        >>> from planet import Session
        >>>
        >>> async def main():
        ...     async with Session() as sess:
        ...         cl = sess.client('features')
        ...         # use client here
        ...
        >>> asyncio.run(main())
        ```
    """

    def __init__(self,
                 session: Session,
                 base_url: Optional[str] = None) -> None:
        """
        Parameters:
            session: Open session connected to server.
            base_url: The base URL to use. Defaults to the Features
                API base url at api.planet.com.
        """
        self._session = session

        self._base_url = base_url or BASE_URL
        if self._base_url.endswith('/'):
            self._base_url = self._base_url[:-1]

    def _call_sync(self, f: Awaitable[T]) -> T:
        """block on an async function call, using the call_sync method of the session"""
        return self._session._call_sync(f)

    async def list_collections(self, limit: int = 0) -> AsyncIterator[dict]:
        """
        List the feature collections you have access to.

        Example:

        ```
        collections = await client.list_collections()
        async for collection in collections:
            print(collection)
        ```
        """

        class _CollectionsPager(Paged):
            """Navigates pages of messages about collections."""
            ITEMS_KEY = 'collections'
            LINKS_KEY = 'links'

            def _next_link(self, page):
                next_link = False

                # TODO: Build this into Paged.__next_link directly, other (unimplemented) APIs have a similar page structure
                for link in page[self.LINKS_KEY]:
                    if "rel" in link and link["rel"] == self.NEXT_KEY:
                        next_link = link["href"]
                        LOGGER.debug(f'next: {next_link}')
                if not next_link:
                    LOGGER.debug('end of the pages')
                return next_link

        url = f'{self._base_url}/collections'

        response = await self._session.request(method='GET', url=url)
        async for col in _CollectionsPager(response,
                                           self._session.request,
                                           limit=limit):
            yield col

    async def get_collection(self, collection_id: str) -> dict:
        """
        Get collection metadata.

        note: when looking up or adding features to a collection,
        you can pass the collection ID directly to the add_features
        or list_features methods. It is not necessary to look up
        the collection metadata first.
        """
        url = f'{self._base_url}/collections/{collection_id}'
        response = await self._session.request(method='GET', url=url)
        return response.json()

    async def list_items(
        self,
        collection_id: str,
        limit: int = 10,
    ) -> AsyncIterator[Feature]:
        """
        List features in `collection_id`.

        Returns an iterator of Features, which are `dict` instances with a
        convenience method/property for getting a feature reference (`.ref`).
        The reference can be used in the Data, Orders and Subscriptions
        APIs. Within the Python SDK, the entire Feature can generally be
        passed to functions that accept a geometry.

        example:

        ```
        results = await client.list_items(collection_id)
        async for feature in results:
            print(feature.ref)
            print(feature["id"])
            print(feature["geometry"])
        ```
        """

        class _FeaturesPager(Paged):
            """Navigates pages of messages about features."""
            ITEMS_KEY = 'features'
            LINKS_KEY = 'links'

            def _next_link(self, page):
                next_link = False

                # TODO: Build this into Paged.__next_link directly, other (unimplemented) APIs have a similar page structure
                for link in page[self.LINKS_KEY]:
                    if "rel" in link and link["rel"] == self.NEXT_KEY:
                        next_link = link["href"]
                        LOGGER.debug(f'next: {next_link}')
                if not next_link:
                    LOGGER.debug('end of the pages')
                return next_link

        url = f'{self._base_url}/collections/{collection_id}/items'

        resp = await self._session.request(method='GET', url=url)
        async for feat in _FeaturesPager(resp,
                                         self._session.request,
                                         limit=limit):
            yield Feature(**feat)

    async def get_item(self, collection_id: str, feature_id: str) -> Feature:
        """
        Return metadata for a single feature in a collection
        """
        url = f'{self._base_url}/collections/{collection_id}/items/{feature_id}'
        response = await self._session.request(method='GET', url=url)
        return Feature(**response.json())

    async def create_collection(self,
                                title: str,
                                description: Optional[str] = None) -> str:
        """
        Create a new collection with the given title and description, returning the collection id.

        Example:

        ```
        collection_id = await features_client.create_collection(
          title="My Collection",
          description="a test collection"
        )
        ```
        """
        url = f'{self._base_url}/collections'
        body = {"title": title}

        if description:
            body["description"] = description

        resp = await self._session.request(method='POST', url=url, json=body)

        return resp.json()["id"]

    async def add_items(self,
                        collection_id: str,
                        feature: Union[dict, GeoInterface],
                        property_id: Optional[str] = None) -> list[str]:
        """
        Add a Feature or FeatureCollection to the collection given by `collection_id`.
        Returns a list of feature references.

        Features API only accepts Polygon and MultiPolygon.

        collection_id: the collection to add the feature to
        feature: a dict containing a geojson Feature or FeatureCollection, or an
          instance of a class that implements __geo_interface__ (e.g. a Shapely or
          GeoPandas geometry object)
        property_id (optional): the name of a property in the `properties` block of
          the supplied feature(s). The value will become the feature's id.

        When the feature is added to the collection, it will be given an ID.
        The ID can be overriden by providing an `id` in the feature's properties.
        Title and description can also be set this way using the `title` and
        `description` properties.

        Example:

        ```
        new_features = await features_client.add_features(
          collection_id="my-collection",
          feature=feature,
        )
        ```

        note: if a geojson Geometry is supplied, it will be converted
        to a Feature. However, we recommend doing this conversion yourself
        so that you can set a title and description property.

        The return value is always an iterator, even if you only upload one
        feature.
        """
        if isinstance(feature, GeoInterface):
            # we expect __geo_interface__ to return a Geometry (not a Feature), so
            # we're using the name `feature` liberally. We'll convert to a feature
            # at the next step.
            feature = feature.__geo_interface__

        # convert a geojson geometry into geojson feature
        if feature.get("type",
                       "").lower() not in ["feature", "featurecollection"]:
            feature = {"type": "Feature", "geometry": feature}

        url = f'{self._base_url}/collections/{collection_id}/items'
        params: dict[str, Any] = {}
        if property_id:
            params["property_id"] = property_id

        resp = await self._session.request(method='POST',
                                           url=url,
                                           json=feature,
                                           params=params)
        return list(resp.json())



# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Functionality for interacting with the data api"""
from pathlib import Path
from typing import Any, Callable, Dict, Iterator, List, Optional, Union

from planet.models import GeojsonLike

from ..http import Session

from planet.clients import DataClient

LIST_SORT_DEFAULT = 'created desc'
LIST_SEARCH_TYPE_DEFAULT = 'any'

WAIT_DELAY = 5
WAIT_MAX_ATTEMPTS = 200


class DataAPI:
    """Data API client"""

    _client: DataClient

    def __init__(self, session: Session, base_url: Optional[str] = None):
        """
        Parameters:
            session: Open session connected to server.
            base_url: The base URL to use. Defaults to production data API
                base url.
        """
        self._client = DataClient(session, base_url)

    def search(
        self,
        item_types: List[str],
        search_filter: Optional[Dict] = None,
        name: Optional[str] = None,
        sort: Optional[str] = None,
        limit: int = 100,
        geometry: Optional[GeojsonLike] = None,
    ) -> Iterator[Dict]:
        """
        Search for items

        Example:

        ```python
        pl = Planet()
        for item in pl.data.search(['PSScene'], limit=5):
            print(item)
        ```

        Parameters:
            item_types: The item types to include in the search.
            search_filter: Structured search criteria to apply. If None,
                no search criteria is applied.
            sort: Field and direction to order results by. Valid options are
            given in SEARCH_SORT.
            name: The name of the saved search.
            limit: Maximum number of results to return. When set to 0, no
                maximum is applied.
            geometry: GeoJSON, a feature reference or a list of feature
                references
        """

        results = self._client.search(item_types,
                                      search_filter,
                                      name,
                                      sort,
                                      limit,
                                      geometry)

        try:
            while True:
                yield self._client._call_sync(results.__anext__())
        except StopAsyncIteration:
            pass

    def create_search(
        self,
        item_types: List[str],
        search_filter: Dict,
        name: str,
        enable_email: bool = False,
        geometry: Optional[Union[Dict, str]] = None,
    ) -> Dict:
        """Create a new saved structured item search.

        To filter to items you have access to download which are of standard
        (aka not test) quality, use the following:

        ```python
        >>> from planet import data_filter
        >>> data_filter.and_filter([
        ...     data_filter.permission_filter(),
        ...     data_filter.std_quality_filter()
        >>> ])

        ```

        To avoid filtering out any imagery, supply a blank AndFilter, which can
        be created with `data_filter.and_filter([])`.


        Parameters:
            item_types: The item types to include in the search.
            search_filter: Structured search criteria.
            name: The name of the saved search.
            enable_email: Send a daily email when new results are added.
            geometry: A feature reference or a GeoJSON

        Returns:
            Description of the saved search.

        Raises:
            planet.exceptions.APIError: On API error.
        """
        return self._client._call_sync(
            self._client.create_search(item_types,
                                       search_filter,
                                       name,
                                       enable_email,
                                       geometry))

    def update_search(
            self,
            search_id: str,
            item_types: List[str],
            search_filter: Dict[str, Any],
            name: str,
            enable_email: bool = False,
            geometry: Optional[Union[dict, str]] = None) -> Dict[str, Any]:
        """Update an existing saved search.

        Parameters:
            search_id: Saved search identifier.
            item_types: The item types to include in the search.
            search_filter: Structured search criteria.
            name: The name of the saved search.
            enable_email: Send a daily email when new results are added.
            geometry: A feature reference or a GeoJSON

        Returns:
            Description of the saved search.
        """
        return self._client._call_sync(
            self._client.update_search(search_id,
                                       item_types,
                                       search_filter,
                                       name,
                                       enable_email,
                                       geometry))

    def list_searches(self,
                      sort: str = LIST_SORT_DEFAULT,
                      search_type: str = LIST_SEARCH_TYPE_DEFAULT,
                      limit: int = 100) -> Iterator[Dict[str, Any]]:
        """Iterate through list of searches available to the user.

        Parameters:
            sort: Field and direction to order results by.
            search_type: Filter to specified search type.
            limit: Maximum number of results to return. When set to 0, no
                maximum is applied.

        Yields:
            Description of a search.

        Raises:
            planet.exceptions.APIError: On API error.
            planet.exceptions.ClientError: If sort or search_type are not
                valid.
        """
        results = self._client.list_searches(sort, search_type, limit)

        try:
            while True:
                yield self._client._call_sync(results.__anext__())
        except StopAsyncIteration:
            pass

    def delete_search(self, search_id: str):
        """Delete an existing saved search.

        Parameters:
            search_id: Saved search identifier.

        Raises:
            planet.exceptions.APIError: On API error.
        """
        return self._client._call_sync(self._client.delete_search(search_id))

    def get_search(self, search_id: str) -> Dict:
        """Get a saved search by id.

        Parameters:
            search_id: Stored search identifier.

        Returns:
            Saved search details.

        Raises:
            planet.exceptions.APIError: On API error.
        """
        return self._client._call_sync(self._client.get_search(search_id))

    def run_search(self,
                   search_id: str,
                   sort: Optional[str] = None,
                   limit: int = 100) -> Iterator[Dict[str, Any]]:
        """Iterate over results from a saved search.

        Note:
            The name of this method is based on the API's method name. This
            method provides iteration over results, it does not get a
            single result description or return a list of descriptions.

        Parameters:
            search_id: Stored search identifier.
            sort: Field and direction to order results by. Valid options are
            given in SEARCH_SORT.
            limit: Maximum number of results to return. When set to 0, no
                maximum is applied.

        Yields:
            Description of an item.

        Raises:
            planet.exceptions.APIError: On API error.
            planet.exceptions.ClientError: If search_id or sort is not valid.
        """

        results = self._client.run_search(search_id, sort, limit)

        try:
            while True:
                yield self._client._call_sync(results.__anext__())
        except StopAsyncIteration:
            pass

    def get_stats(self,
                  item_types: List[str],
                  search_filter: Dict[str, Any],
                  interval: str) -> Dict[str, Any]:
        """Get item search statistics.

        Parameters:
            item_types: The item types to include in the search.
            search_filter: Structured search criteria.
            interval: The size of the histogram date buckets.

        Returns:
            A full JSON description of the returned statistics result
            histogram.

        Raises:
            planet.exceptions.APIError: On API error.
            planet.exceptions.ClientError: If interval is not valid.
        """
        return self._client._call_sync(
            self._client.get_stats(item_types, search_filter, interval))

    def list_item_assets(self, item_type_id: str,
                         item_id: str) -> Dict[str, Any]:
        """List all assets available for an item.

        An asset describes a product that can be derived from an item's source
        data, and can be used for various analytic, visual or other purposes.
        These are referred to as asset_types.

        Parameters:
            item_type_id: Item type identifier.
            item_id: Item identifier.

        Returns:
            Descriptions of available assets as a dictionary with asset_type_id
            as keys and asset description as value.

        Raises:
            planet.exceptions.APIError: On API error.
        """
        return self._client._call_sync(
            self._client.list_item_assets(item_type_id, item_id))

    def get_asset(self, item_type_id: str, item_id: str,
                  asset_type_id: str) -> Dict[str, Any]:
        """Get an item asset description.

        Parameters:
            item_type_id: Item type identifier.
            item_id: Item identifier.
            asset_type_id: Asset type identifier.

        Returns:
            Description of the asset.

        Raises:
            planet.exceptions.APIError: On API error.
            planet.exceptions.ClientError: If asset type identifier is not
            valid.
        """
        return self._client._call_sync(
            self._client.get_asset(item_type_id, item_id, asset_type_id))

    def activate_asset(self, asset: Dict[str, Any]):
        """Activate an item asset.

        Parameters:
            asset: Description of the asset. Obtained from get_asset().

        Raises:
            planet.exceptions.APIError: On API error.
            planet.exceptions.ClientError: If asset description is not
            valid.
        """
        return self._client._call_sync(self._client.activate_asset(asset))

    def wait_asset(
            self,
            asset: dict,
            delay: int = WAIT_DELAY,
            max_attempts: int = WAIT_MAX_ATTEMPTS,
            callback: Optional[Callable[[str],
                                        None]] = None) -> Dict[Any, Any]:
        """Wait for an item asset to be active.

        Prior to waiting for the asset to be active, be sure to activate the
        asset with activate_asset().

        Parameters:
            asset: Description of the asset. Obtained from get_asset().
            delay: Time (in seconds) between polls.
            max_attempts: Maximum number of polls. When set to 0, no limit
                is applied.
            callback: Function that handles state progress updates.

        Returns:
            Last received description of the asset.

        Raises:
            planet.exceptions.APIError: On API error.
            planet.exceptions.ClientError: If asset_type_id is not valid or is
                not available or if the maximum number of attempts is reached
                before the asset is active.
        """
        return self._client._call_sync(
            self._client.wait_asset(asset, delay, max_attempts, callback))

    def download_asset(self,
                       asset: dict,
                       filename: Optional[str] = None,
                       directory: Path = Path('.'),
                       overwrite: bool = False,
                       progress_bar: bool = True) -> Path:
        """Download an asset.

        The asset must be active before it can be downloaded. This can be
        achieved with activate_asset() followed by wait_asset().

        If overwrite is False and the file already exists, download will be
        skipped and the file path will be returned as usual.

        Parameters:
            asset: Description of the asset. Obtained from get_asset() or
                wait_asset().
            filename: Custom name to assign to downloaded file.
            directory: Base directory for file download.
            overwrite: Overwrite any existing files.
            progress_bar: Show progress bar during download.

        Returns:
            Path to downloaded file.

        Raises:
            planet.exceptions.APIError: On API error.
            planet.exceptions.ClientError: If asset is not active or asset
            description is not valid.
        """
        return self._client._call_sync(
            self._client.download_asset(asset,
                                        filename,
                                        directory,
                                        overwrite,
                                        progress_bar))

    @staticmethod
    def validate_checksum(asset: Dict[str, Any], filename: Path):
        """Validate checksum of downloaded file

        Compares checksum calculated from the file against the value provided
        in the asset.


        Parameters:
            asset: Description of the asset. Obtained from get_asset() or
                wait_asset().
            filename: Full path to downloaded file.

        Raises:
            planet.exceptions.ClientError: If the file does not exist or if
                checksums do not match.
        """
        return DataClient.validate_checksum(asset, filename)



# Copyright 2017 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import math
from unittest.mock import MagicMock
import os
from pathlib import Path
import re

import pytest

from planet import models
from planet.exceptions import PagingError

LOGGER = logging.getLogger(__name__)


def test_StreamingBody_name_filename():
    r = MagicMock(name='response')
    r.url = 'https://planet.com/path/to/example.tif?foo=f6f1'
    r.headers = {
        'date': 'Thu, 14 Feb 2019 16:13:26 GMT',
        'last-modified': 'Wed, 22 Nov 2017 17:22:31 GMT',
        'accept-ranges': 'bytes',
        'content-type': 'image/tiff',
        'content-length': '57350256',
        'content-disposition': 'attachment; filename="open_california.tif"'
    }
    body = models.StreamingBody(r)
    assert body.name == 'open_california.tif'


def test_StreamingBody_name_url():
    r = MagicMock(name='response')
    r.url = 'https://planet.com/path/to/example.tif?foo=f6f1'
    r.headers = {
        'date': 'Thu, 14 Feb 2019 16:13:26 GMT',
        'last-modified': 'Wed, 22 Nov 2017 17:22:31 GMT',
        'accept-ranges': 'bytes',
        'content-type': 'image/tiff',
        'content-length': '57350256',
    }
    body = models.StreamingBody(r)

    assert body.name == 'example.tif'


def test_StreamingBody_name_content():
    r = MagicMock(name='response')
    r.url = 'https://planet.com/path/to/noname/'
    r.headers = {
        'date': 'Thu, 14 Feb 2019 16:13:26 GMT',
        'last-modified': 'Wed, 22 Nov 2017 17:22:31 GMT',
        'accept-ranges': 'bytes',
        'content-type': 'image/tiff',
        'content-length': '57350256',
    }
    body = models.StreamingBody(r)

    assert body.name.startswith('planet-')
    assert (body.name.endswith('.tiff') or body.name.endswith('.tif'))


NO_NAME_HEADERS = {
    'date': 'Thu, 14 Feb 2019 16:13:26 GMT',
    'last-modified': 'Wed, 22 Nov 2017 17:22:31 GMT',
    'accept-ranges': 'bytes',
    'content-type': 'image/tiff',
    'content-length': '57350256'
}
OPEN_CALIFORNIA_HEADERS = {
    'date': 'Thu, 14 Feb 2019 16:13:26 GMT',
    'last-modified': 'Wed, 22 Nov 2017 17:22:31 GMT',
    'accept-ranges': 'bytes',
    'content-type': 'image/tiff',
    'content-length': '57350256',
    'content-disposition': 'attachment; filename="open_california.tif"'
}


@pytest.mark.parametrize('headers,expected',
                         [(OPEN_CALIFORNIA_HEADERS, 'open_california.tif'),
                          (NO_NAME_HEADERS, None),
                          ({}, None)])  # yapf: disable
def test__get_filename_from_headers(headers, expected):
    assert models._get_filename_from_headers(headers) == expected


@pytest.mark.parametrize(
    'url,expected',
    [
        ('https://planet.com/', None),
        ('https://planet.com/path/to/', None),
        ('https://planet.com/path/to/example.tif', 'example.tif'),
        ('https://planet.com/path/to/example.tif?foo=f6f1&bar=baz',
         'example.tif'),
        ('https://planet.com/path/to/example.tif?foo=f6f1#quux',
         'example.tif'),
    ])
def test__get_filename_from_url(url, expected):
    assert models._get_filename_from_url(url) == expected


@pytest.mark.parametrize(
    'content_type,check',
    [
        (None,
         lambda x: re.match(r'^planet-[a-z0-9]{8}$', x, re.I) is not None),
        ('image/tiff', lambda x: x.endswith(('.tif', '.tiff'))),
    ])
def test__get_random_filename(content_type, check):
    assert check(models._get_random_filename(content_type))


@pytest.mark.anyio
async def test_StreamingBody_write_img(tmpdir, open_test_img):

    async def _aiter_bytes():
        data = open_test_img.read()
        v = memoryview(data)

        chunksize = 100
        for i in range(math.ceil(len(v) / (chunksize))):
            yield v[i * chunksize:min((i + 1) * chunksize, len(v))]

    r = MagicMock(name='response')
    r.aiter_bytes = _aiter_bytes
    r.num_bytes_downloaded = 0
    r.headers['Content-Length'] = 527
    body = models.StreamingBody(r)

    filename = Path(tmpdir) / 'test.tif'
    await body.write(filename, progress_bar=False)

    assert os.path.isfile(filename)
    assert os.stat(filename).st_size == 527


@pytest.mark.anyio
async def test_Paged_iterator():
    resp = MagicMock(name='response')
    resp.json = lambda: {'_links': {'next': 'blah'}, 'items': [1, 2]}

    async def get_response(url, method):
        resp = MagicMock(name='response')
        resp.json = lambda: {'_links': {}, 'items': [3, 4]}
        return resp

    paged = models.Paged(resp, get_response)
    assert [1, 2, 3, 4] == [i async for i in paged]


@pytest.mark.anyio
@pytest.mark.parametrize('limit, expected', [(0, [1, 2, 3, 4]), (1, [1])])
async def test_Paged_limit(limit, expected):
    resp = MagicMock(name='response')
    resp.json = lambda: {'_links': {'next': 'blah'}, 'items': [1, 2]}

    async def get_response(url, method):
        resp = MagicMock(name='response')
        resp.json = lambda: {'_links': {}, 'items': [3, 4]}
        return resp

    paged = models.Paged(resp, get_response, limit=limit)
    assert [i async for i in paged] == expected


@pytest.mark.anyio
async def test_Paged_break_page_cycle():
    """Check that we break out of a page cycle."""
    resp = MagicMock(name='response')
    resp.json = lambda: {'_links': {'next': 'blah'}, 'items': [1, 2]}

    async def get_response(url, method):
        return resp

    paged = models.Paged(resp, get_response, limit=None)

    with pytest.raises(PagingError):
        [item async for item in paged]



import tempfile
import json
import pytest

import respx
from click.testing import CliRunner

from planet.cli import cli
from tests.integration.test_features_api import TEST_COLLECTION_1, TEST_COLLECTION_LIST, TEST_FEAT, TEST_GEOM, TEST_URL, list_collections_response, list_features_response, mock_response, to_collection_model


def invoke(*args):

    runner = CliRunner()
    opts = ["--base-url", TEST_URL]
    args = ['features'] + opts + [arg for arg in args]

    result = runner.invoke(cli.main, args=args)
    assert result.exit_code == 0, result.output
    return json.loads(result.output)


@respx.mock
def test_collection_list():

    collections_url = f'{TEST_URL}/collections'
    mock_response(collections_url,
                  list_collections_response(TEST_COLLECTION_LIST))

    def assertf(resp):
        assert resp[0]["id"] == "collection1"
        assert resp[1]["id"] == "collection2"

    assertf(invoke("collections", "list"))


@respx.mock
def test_get_collection():
    id = TEST_COLLECTION_1["id"]
    collection_url = f'{TEST_URL}/collections/{id}'

    mock_response(collection_url, to_collection_model(TEST_COLLECTION_1))

    def assertf(resp):
        assert resp["id"] == "collection1"
        assert resp["title"] == "Collection 1"

    assertf(invoke("collections", "get", id))


@respx.mock
def test_create_collection():

    collection_url = f'{TEST_URL}/collections'

    mock_response(collection_url,
                  to_collection_model(TEST_COLLECTION_1),
                  method='post')

    def assertf(resp):
        # the return value is simply the id.
        assert resp == "collection1"

    assertf(
        invoke("collections",
               "create",
               "--title",
               TEST_COLLECTION_1["title"],
               "--description",
               TEST_COLLECTION_1["description"]))

    req_body = json.loads(respx.calls[0].request.content)
    assert req_body["title"] == "Collection 1"
    assert req_body["description"] == "test collection 1"


@respx.mock
def test_list_items():
    collection_id = "test"
    items_url = f'{TEST_URL}/collections/{collection_id}/items'

    mock_response(items_url,
                  list_features_response(collection_id, num_features=3))

    def assertf(resp):
        assert resp[0]["id"] == "0"
        assert resp[1]["id"] == "1"

    assertf(invoke("items", "list", collection_id))


@respx.mock
@pytest.mark.parametrize("feature, expected_body", [
    (TEST_FEAT, TEST_GEOM),
    (TEST_GEOM, TEST_GEOM),
])
def test_add_items(feature, expected_body):
    """test adding a feature with the CLI
    cases:
    * a geojson Feature
    * a geojson Geometry
    """
    collection_id = "test"
    items_url = f'{TEST_URL}/collections/{collection_id}/items'

    # mock a feature ref return
    feat_resp = ["pl:features/my/test/test1"]
    mock_response(items_url, feat_resp, method="post")

    def assertf(resp):
        # check that mocked response is returned in full
        assert resp == feat_resp

    with tempfile.NamedTemporaryFile('w+') as file:
        json.dump(feature, file)
        file.flush()

        assertf(invoke("items", "add", collection_id, file.name))

        # check request body. In all test cases, the request body
        # should be a geojson Feature.
        req_body = json.loads(respx.calls[0].request.content)
        assert req_body["type"] == "Feature"
        assert req_body["geometry"] == expected_body



# Copyright 2020 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
from contextlib import nullcontext as does_not_raise
import copy
import json
import hashlib
from http import HTTPStatus
import logging
import math
import os
from pathlib import Path
from unittest.mock import call, create_autospec

import httpx
import pytest
import respx

from planet import OrdersClient, exceptions, reporting
from planet.clients.orders import OrderStates
from planet.sync import Planet

TEST_URL = 'http://www.MockNotRealURL.com/api/path'
TEST_BULK_CANCEL_URL = f'{TEST_URL}/bulk/orders/v2/cancel'
TEST_DOWNLOAD_URL = f'{TEST_URL}/download'
TEST_DOWNLOAD_ACTUAL_URL = f'{TEST_URL}/download_actual'
TEST_ORDERS_URL = f'{TEST_URL}/orders/v2'
TEST_STATS_URL = f'{TEST_URL}/stats/orders/v2'

LOGGER = logging.getLogger(__name__)


@pytest.fixture
def downloaded_content():
    return {'key': 'downloaded_file'}


@pytest.fixture
def original_content():
    return {'key': 'original_file'}


@pytest.fixture
def create_download_mock(downloaded_content, order_description, oid):
    """
    Mock an HTTP response for download.
    """

    def f():
        # Create mock HTTP response
        order_description['state'] = 'success'
        dl_url = TEST_DOWNLOAD_URL + '/1?token=IAmAToken'
        order_description['_links']['results'] = [
            {
                'location': dl_url, 'name': 'file.json'
            },
        ]

        get_url = f'{TEST_ORDERS_URL}/{oid}'
        mock_resp = httpx.Response(HTTPStatus.OK, json=order_description)
        respx.get(get_url).return_value = mock_resp

        mock_resp = httpx.Response(
            HTTPStatus.OK,
            json=downloaded_content,
            headers={
                'Content-Type': 'application/json',
                'Content-Disposition': 'attachment; filename="file.json"'
            })
        respx.get(dl_url).return_value = mock_resp

    return f


def test_OrderStates_reached():
    assert not OrderStates.reached('running', 'queued')
    assert OrderStates.reached('running', 'running')
    assert OrderStates.reached('running', 'failed')


def test_OrderStates_passed():
    assert not OrderStates.reached('running', 'queued')
    assert not OrderStates.passed('running', 'running')
    assert OrderStates.passed('running', 'success')


@respx.mock
@pytest.mark.anyio
async def test_list_orders_basic(order_descriptions, session):
    next_page_url = TEST_ORDERS_URL + 'blob/?page_marker=IAmATest'

    order1, order2, order3 = order_descriptions

    page1_response = {
        "_links": {
            "_self": "string", "next": next_page_url
        },
        "orders": [order1, order2]
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=page1_response)
    respx.get(TEST_ORDERS_URL).return_value = mock_resp

    page2_response = {"_links": {"_self": next_page_url}, "orders": [order3]}
    mock_resp2 = httpx.Response(HTTPStatus.OK, json=page2_response)
    respx.get(next_page_url).return_value = mock_resp2

    cl = OrdersClient(session, base_url=TEST_URL)
    assert order_descriptions == [o async for o in cl.list_orders()]


@respx.mock
def test_list_orders_basic_sync(order_descriptions, session):
    next_page_url = TEST_ORDERS_URL + 'blob/?page_marker=IAmATest'

    order1, order2, order3 = order_descriptions

    page1_response = {
        "_links": {
            "_self": "string", "next": next_page_url
        },
        "orders": [order1, order2]
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=page1_response)
    respx.get(TEST_ORDERS_URL).return_value = mock_resp

    page2_response = {"_links": {"_self": next_page_url}, "orders": [order3]}
    mock_resp2 = httpx.Response(HTTPStatus.OK, json=page2_response)
    respx.get(next_page_url).return_value = mock_resp2

    pl = Planet()
    pl.orders._client._base_url = TEST_URL
    assert order_descriptions == list(pl.orders.list_orders())


@respx.mock
@pytest.mark.anyio
async def test_list_orders_filtering_and_sorting(order_descriptions, session):
    list_url = TEST_ORDERS_URL + '?source_type=all&state=failed&name=my_order_xyz&name__contains=xyz&created_on=2018-02-12T00:00:00Z/..&last_modified=../2018-03-18T12:31:12Z&hosting=true&sort_by=name DESC'

    order1, order2, _ = order_descriptions

    page1_response = {
        "_links": {
            "_self": "string"
        }, "orders": [order1, order2]
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=page1_response)
    respx.get(list_url).return_value = mock_resp

    cl = OrdersClient(session, base_url=TEST_URL)

    # if the value of each arg doesn't get sent as a url parameter,
    # the mock will fail and this test will fail
    assert [order1, order2] == [
        o async for o in cl.list_orders(
            state='failed', name='my_order_xyz', name__contains='xyz',
            created_on='2018-02-12T00:00:00Z/..',
            last_modified='../2018-03-18T12:31:12Z', hosting=True,
            sort_by='name DESC')
    ]


@respx.mock
def test_list_orders_state_success_sync(order_descriptions, session):
    list_url = TEST_ORDERS_URL + '?source_type=all&state=failed'

    order1, order2, _ = order_descriptions

    page1_response = {
        "_links": {
            "_self": "string"
        }, "orders": [order1, order2]
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=page1_response)
    respx.get(list_url).return_value = mock_resp

    pl = Planet()
    pl.orders._client._base_url = TEST_URL

    # if the value of state doesn't get sent as a url parameter,
    # the mock will fail and this test will fail
    assert [order1, order2] == list(pl.orders.list_orders(state='failed'))


@pytest.mark.anyio
async def test_list_orders_state_invalid(session):
    cl = OrdersClient(session, base_url=TEST_URL)

    with pytest.raises(exceptions.ClientError):
        [o async for o in cl.list_orders(state='invalidstate')]


def test_list_orders_state_invalid_sync(session):
    pl = Planet()
    pl.orders._client._base_url = TEST_URL

    with pytest.raises(exceptions.ClientError):
        list(pl.orders.list_orders(state='invalidstate'))


@respx.mock
@pytest.mark.anyio
@pytest.mark.parametrize("limit,limited_list_length", [(None, 100), (0, 102),
                                                       (1, 1)])
async def test_list_orders_limit(order_descriptions,
                                 session,
                                 limit,
                                 limited_list_length):
    nono_page_url = None

    # Creating 102 (3x34) order descriptions
    long_order_descriptions = order_descriptions * 34

    all_orders = {}
    for x in range(1, len(long_order_descriptions) + 1):
        all_orders["order{0}".format(x)] = long_order_descriptions[x - 1]

    page1_response = {
        "_links": {
            "_self": "string", "next": nono_page_url
        },
        "orders": [
            all_orders['order%s' % num]
            for num in range(1, limited_list_length + 1)
        ]
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=page1_response)
    respx.get(TEST_ORDERS_URL).return_value = mock_resp

    cl = OrdersClient(session, base_url=TEST_URL)

    assert len([o async for o in cl.list_orders(limit=limit)
                ]) == limited_list_length


@respx.mock
@pytest.mark.anyio
async def test_create_order_basic(oid,
                                  order_description,
                                  order_request,
                                  session):
    route = respx.post(TEST_ORDERS_URL)
    route.return_value = httpx.Response(HTTPStatus.OK, json=order_description)

    cl = OrdersClient(session, base_url=TEST_URL)
    order = await cl.create_order(order_request)

    assert order == order_description

    assert json.loads(route.calls.last.request.content) == order_request


@respx.mock
def test_create_order_basic_sync(oid,
                                 order_description,
                                 order_request,
                                 session):
    route = respx.post(TEST_ORDERS_URL)
    route.return_value = httpx.Response(HTTPStatus.OK, json=order_description)

    pl = Planet()
    pl.orders._client._base_url = TEST_URL
    order = pl.orders.create_order(order_request)

    assert order == order_description

    assert json.loads(route.calls.last.request.content) == order_request


@respx.mock
@pytest.mark.anyio
async def test_create_order_bad_item_type(order_request, session):
    resp = {
        "field": {
            "Products": [{
                "message": "Bad item type 'invalid' for bundle type 'analytic'"
            }]
        },
        "general": [{
            "message": "Unable to accept order"
        }]
    }
    mock_resp = httpx.Response(400, json=resp)
    respx.post(TEST_ORDERS_URL).return_value = mock_resp
    order_request['products'][0]['item_type'] = 'invalid'
    cl = OrdersClient(session, base_url=TEST_URL)

    with pytest.raises(exceptions.BadQuery):
        await cl.create_order(order_request)


@respx.mock
@pytest.mark.anyio
async def test_create_order_item_id_does_not_exist(order_request, session):
    resp = {
        "field": {
            "Details": [{
                "message": ("Item ID 4500474_2133707_2021-05-20_2419 / " +
                            "Item Type PSScene3Band doesn't exist")
            }]
        },
        "general": [{
            "message": "Unable to accept order"
        }]
    }
    mock_resp = httpx.Response(400, json=resp)
    respx.post(TEST_ORDERS_URL).return_value = mock_resp
    order_request['products'][0]['item_ids'] = \
        '4500474_2133707_2021-05-20_2419'
    cl = OrdersClient(session, base_url=TEST_URL)

    with pytest.raises(exceptions.BadQuery):
        await cl.create_order(order_request)


@respx.mock
@pytest.mark.anyio
async def test_get_order(oid, order_description, session):
    get_url = f'{TEST_ORDERS_URL}/{oid}'
    mock_resp = httpx.Response(HTTPStatus.OK, json=order_description)
    respx.get(get_url).return_value = mock_resp

    cl = OrdersClient(session, base_url=TEST_URL)
    order = await cl.get_order(oid)
    assert order_description == order


@respx.mock
def test_get_order_sync(oid, order_description, session):
    get_url = f'{TEST_ORDERS_URL}/{oid}'
    mock_resp = httpx.Response(HTTPStatus.OK, json=order_description)
    respx.get(get_url).return_value = mock_resp

    pl = Planet()
    pl.orders._client._base_url = TEST_URL
    order = pl.orders.get_order(oid)
    assert order_description == order


@pytest.mark.anyio
async def test_get_order_invalid_id(session):
    cl = OrdersClient(session, base_url=TEST_URL)
    with pytest.raises(exceptions.ClientError):
        await cl.get_order('-')


@respx.mock
@pytest.mark.anyio
async def test_get_order_id_doesnt_exist(oid, session):
    get_url = f'{TEST_ORDERS_URL}/{oid}'

    resp = {"message": f'Could not load order ID: {oid}'}
    mock_resp = httpx.Response(404, json=resp)
    respx.get(get_url).return_value = mock_resp

    cl = OrdersClient(session, base_url=TEST_URL)

    with pytest.raises(exceptions.MissingResource):
        await cl.get_order(oid)


@respx.mock
@pytest.mark.anyio
async def test_cancel_order(oid, order_description, session):
    cancel_url = f'{TEST_ORDERS_URL}/{oid}'
    order_description['state'] = 'cancelled'
    mock_resp = httpx.Response(HTTPStatus.OK, json=order_description)
    example_resp = mock_resp.json()
    respx.put(cancel_url).return_value = mock_resp

    cl = OrdersClient(session, base_url=TEST_URL)
    json_resp = await cl.cancel_order(oid)
    assert json_resp == example_resp


@respx.mock
def test_cancel_order_sync(oid, order_description, session):
    cancel_url = f'{TEST_ORDERS_URL}/{oid}'
    order_description['state'] = 'cancelled'
    mock_resp = httpx.Response(HTTPStatus.OK, json=order_description)
    example_resp = mock_resp.json()
    respx.put(cancel_url).return_value = mock_resp

    pl = Planet()
    pl.orders._client._base_url = TEST_URL
    json_resp = pl.orders.cancel_order(oid)
    assert json_resp == example_resp


@pytest.mark.anyio
async def test_cancel_order_invalid_id(session):
    cl = OrdersClient(session, base_url=TEST_URL)
    with pytest.raises(exceptions.ClientError):
        await cl.cancel_order('invalid_order_id')


@respx.mock
@pytest.mark.anyio
async def test_cancel_order_id_doesnt_exist(oid, session):
    cancel_url = f'{TEST_ORDERS_URL}/{oid}'

    resp = {"message": f'No such order ID: {oid}.'}
    mock_resp = httpx.Response(404, json=resp)
    respx.put(cancel_url).return_value = mock_resp

    cl = OrdersClient(session, base_url=TEST_URL)

    with pytest.raises(exceptions.MissingResource):
        await cl.cancel_order(oid)


@respx.mock
@pytest.mark.anyio
async def test_cancel_order_id_cannot_be_cancelled(oid, session):
    cancel_url = f'{TEST_ORDERS_URL}/{oid}'

    resp = {"message": 'Order not in a cancellable state'}
    mock_resp = httpx.Response(409, json=resp)
    respx.put(cancel_url).return_value = mock_resp

    cl = OrdersClient(session, base_url=TEST_URL)

    with pytest.raises(exceptions.Conflict):
        await cl.cancel_order(oid)


@respx.mock
@pytest.mark.anyio
async def test_cancel_orders_by_ids(session, oid):
    oid2 = '5ece1dc0-ea81-11eb-837c-acde48001122'
    test_ids = [oid, oid2]
    example_result = {
        "result": {
            "succeeded": {
                "count": 1
            },
            "failed": {
                "count": 1,
                "failures": [{
                    "order_id": oid2,
                    "message": "Order not in a cancellable state",
                }]
            }
        }
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=example_result)
    respx.post(TEST_BULK_CANCEL_URL).return_value = mock_resp

    cl = OrdersClient(session, base_url=TEST_URL)
    res = await cl.cancel_orders(test_ids)

    assert res == example_result

    expected_body = {"order_ids": test_ids}
    actual_body = json.loads(respx.calls.last.request.content)
    assert actual_body == expected_body


@respx.mock
def test_cancel_orders_by_ids_sync(session, oid):
    oid2 = '5ece1dc0-ea81-11eb-837c-acde48001122'
    test_ids = [oid, oid2]
    example_result = {
        "result": {
            "succeeded": {
                "count": 1
            },
            "failed": {
                "count": 1,
                "failures": [{
                    "order_id": oid2,
                    "message": "Order not in a cancellable state",
                }]
            }
        }
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=example_result)
    respx.post(TEST_BULK_CANCEL_URL).return_value = mock_resp

    pl = Planet()
    pl.orders._client._base_url = TEST_URL
    res = pl.orders.cancel_orders(test_ids)

    assert res == example_result

    expected_body = {"order_ids": test_ids}
    actual_body = json.loads(respx.calls.last.request.content)
    assert actual_body == expected_body


@pytest.mark.anyio
async def test_cancel_orders_by_ids_invalid_id(session, oid):
    cl = OrdersClient(session, base_url=TEST_URL)
    with pytest.raises(exceptions.ClientError):
        _ = await cl.cancel_orders([oid, "invalid_oid"])


@respx.mock
@pytest.mark.anyio
async def test_cancel_orders_all(session):
    example_result = {
        "result": {
            "succeeded": {
                "count": 2
            }, "failed": {
                "count": 0, "failures": []
            }
        }
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=example_result)
    respx.post(TEST_BULK_CANCEL_URL).return_value = mock_resp

    cl = OrdersClient(session, base_url=TEST_URL)
    res = await cl.cancel_orders()

    assert res == example_result

    actual_body = json.loads(respx.calls.last.request.content)
    assert actual_body == {}


@respx.mock
def test_cancel_orders_all_sync(session):
    example_result = {
        "result": {
            "succeeded": {
                "count": 2
            }, "failed": {
                "count": 0, "failures": []
            }
        }
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=example_result)
    respx.post(TEST_BULK_CANCEL_URL).return_value = mock_resp

    pl = Planet()
    pl.orders._client._base_url = TEST_URL
    res = pl.orders.cancel_orders()

    assert res == example_result

    actual_body = json.loads(respx.calls.last.request.content)
    assert actual_body == {}


@respx.mock
@pytest.mark.anyio
async def test_wait_default(oid, order_description, session):
    get_url = f'{TEST_ORDERS_URL}/{oid}'

    order_description2 = copy.deepcopy(order_description)
    order_description2['state'] = 'running'
    order_description3 = copy.deepcopy(order_description)
    order_description3['state'] = 'success'

    route = respx.get(get_url)
    route.side_effect = [
        httpx.Response(HTTPStatus.OK, json=order_description),
        httpx.Response(HTTPStatus.OK, json=order_description2),
        httpx.Response(HTTPStatus.OK, json=order_description3)
    ]

    cl = OrdersClient(session, base_url=TEST_URL)
    state = await cl.wait(oid, delay=0)
    assert state == 'success'


@respx.mock
def test_wait_default_sync(oid, order_description, session):
    get_url = f'{TEST_ORDERS_URL}/{oid}'

    order_description2 = copy.deepcopy(order_description)
    order_description2['state'] = 'running'
    order_description3 = copy.deepcopy(order_description)
    order_description3['state'] = 'success'

    route = respx.get(get_url)
    route.side_effect = [
        httpx.Response(HTTPStatus.OK, json=order_description),
        httpx.Response(HTTPStatus.OK, json=order_description2),
        httpx.Response(HTTPStatus.OK, json=order_description3)
    ]

    pl = Planet()
    pl.orders._client._base_url = TEST_URL
    state = pl.orders.wait(oid, delay=0)
    assert state == 'success'


@respx.mock
@pytest.mark.anyio
async def test_wait_callback(oid, order_description, session):
    get_url = f'{TEST_ORDERS_URL}/{oid}'

    order_description2 = copy.deepcopy(order_description)
    order_description2['state'] = 'running'
    order_description3 = copy.deepcopy(order_description)
    order_description3['state'] = 'success'

    route = respx.get(get_url)
    route.side_effect = [
        httpx.Response(HTTPStatus.OK, json=order_description),
        httpx.Response(HTTPStatus.OK, json=order_description2),
        httpx.Response(HTTPStatus.OK, json=order_description3)
    ]

    mock_bar = create_autospec(reporting.StateBar)
    mock_callback = mock_bar.update_state

    cl = OrdersClient(session, base_url=TEST_URL)
    await cl.wait(oid, delay=0, callback=mock_callback)

    # check state was sent to callback as expected
    expected = [call(s) for s in ['queued', 'running', 'success']]
    mock_callback.assert_has_calls(expected)


@respx.mock
def test_wait_callback_sync(oid, order_description, session):
    get_url = f'{TEST_ORDERS_URL}/{oid}'

    order_description2 = copy.deepcopy(order_description)
    order_description2['state'] = 'running'
    order_description3 = copy.deepcopy(order_description)
    order_description3['state'] = 'success'

    route = respx.get(get_url)
    route.side_effect = [
        httpx.Response(HTTPStatus.OK, json=order_description),
        httpx.Response(HTTPStatus.OK, json=order_description2),
        httpx.Response(HTTPStatus.OK, json=order_description3)
    ]

    mock_bar = create_autospec(reporting.StateBar)
    mock_callback = mock_bar.update_state

    pl = Planet()
    pl.orders._client._base_url = TEST_URL
    pl.orders.wait(oid, delay=0, callback=mock_callback)

    # check state was sent to callback as expected
    expected = [call(s) for s in ['queued', 'running', 'success']]
    mock_callback.assert_has_calls(expected)


@respx.mock
@pytest.mark.anyio
async def test_wait_state(oid, order_description, session):
    get_url = f'{TEST_ORDERS_URL}/{oid}'

    order_description2 = copy.deepcopy(order_description)
    order_description2['state'] = 'running'
    order_description3 = copy.deepcopy(order_description)
    order_description3['state'] = 'success'

    route = respx.get(get_url)
    route.side_effect = [
        httpx.Response(HTTPStatus.OK, json=order_description),
        httpx.Response(HTTPStatus.OK, json=order_description2),
        httpx.Response(HTTPStatus.OK, json=order_description3)
    ]

    cl = OrdersClient(session, base_url=TEST_URL)
    state = await cl.wait(oid, state='running', delay=0)
    assert state == 'running'


@respx.mock
@pytest.mark.anyio
async def test_wait_max_attempts_enabled(oid, order_description, session):
    get_url = f'{TEST_ORDERS_URL}/{oid}'

    route = respx.get(get_url)
    route.side_effect = [
        httpx.Response(HTTPStatus.OK, json=order_description),
    ]

    cl = OrdersClient(session, base_url=TEST_URL)
    with pytest.raises(exceptions.ClientError):
        await cl.wait(oid, max_attempts=1, delay=0)


@respx.mock
@pytest.mark.anyio
async def test_wait_max_attempts_disabled(oid, order_description, session):
    get_url = f'{TEST_ORDERS_URL}/{oid}'

    route = respx.get(get_url)
    route.side_effect = [
        httpx.Response(HTTPStatus.OK, json=order_description),
    ] * 300

    cl = OrdersClient(session, base_url=TEST_URL)
    with pytest.raises(RuntimeError):
        # falls off the end of the mocked responses
        await cl.wait(oid, max_attempts=0, delay=0)


@pytest.mark.anyio
async def test_wait_invalid_oid(session):
    cl = OrdersClient(session, base_url=TEST_URL)
    with pytest.raises(exceptions.ClientError):
        await cl.wait("invalid_oid", delay=0)


@pytest.mark.anyio
async def test_wait_invalid_state(oid, session):
    cl = OrdersClient(session, base_url=TEST_URL)
    with pytest.raises(exceptions.ClientError):
        await cl.wait(oid, state="invalid_state", delay=0)


@respx.mock
@pytest.mark.anyio
async def test_aggegated_order_stats(session):
    example_stats = {
        "organization": {
            "queued_orders": 0, "running_orders": 6
        },
        "user": {
            "queued_orders": 0, "running_orders": 0
        }
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=example_stats)
    respx.get(TEST_STATS_URL).return_value = mock_resp

    cl = OrdersClient(session, base_url=TEST_URL)
    res = await cl.aggregated_order_stats()

    assert res == example_stats


@respx.mock
def test_aggegated_order_stats_sync(session):
    example_stats = {
        "organization": {
            "queued_orders": 0, "running_orders": 6
        },
        "user": {
            "queued_orders": 0, "running_orders": 0
        }
    }
    mock_resp = httpx.Response(HTTPStatus.OK, json=example_stats)
    respx.get(TEST_STATS_URL).return_value = mock_resp

    pl = Planet()
    pl.orders._client._base_url = TEST_URL
    res = pl.orders.aggregated_order_stats()

    assert res == example_stats


@respx.mock
@pytest.mark.anyio
async def test_download_asset_md(tmpdir, session):
    dl_url = TEST_DOWNLOAD_URL + '/1?token=IAmAToken'

    md_json = {'key': 'value'}
    md_headers = {
        'Content-Type': 'application/json',
        'Content-Disposition': 'attachment; filename="metadata.json"'
    }

    mock_redirect = httpx.Response(HTTPStatus.FOUND,
                                   headers={
                                       'Location': TEST_DOWNLOAD_ACTUAL_URL,
                                       'Content-Length': '0'
                                   })
    mock_resp = httpx.Response(HTTPStatus.OK, json=md_json, headers=md_headers)

    respx.get(dl_url).return_value = mock_redirect
    respx.get(TEST_DOWNLOAD_ACTUAL_URL).return_value = mock_resp

    cl = OrdersClient(session, base_url=TEST_URL)
    filename = await cl.download_asset(dl_url, directory=str(tmpdir))

    with open(filename) as f:
        assert json.load(f) == {'key': 'value'}

    assert Path(filename).name == 'metadata.json'


@respx.mock
def test_download_asset_md_sync(tmpdir, session):
    dl_url = TEST_DOWNLOAD_URL + '/1?token=IAmAToken'

    md_json = {'key': 'value'}
    md_headers = {
        'Content-Type': 'application/json',
        'Content-Disposition': 'attachment; filename="metadata.json"'
    }

    mock_redirect = httpx.Response(HTTPStatus.FOUND,
                                   headers={
                                       'Location': TEST_DOWNLOAD_ACTUAL_URL,
                                       'Content-Length': '0'
                                   })
    mock_resp = httpx.Response(HTTPStatus.OK, json=md_json, headers=md_headers)

    respx.get(dl_url).return_value = mock_redirect
    respx.get(TEST_DOWNLOAD_ACTUAL_URL).return_value = mock_resp

    pl = Planet()
    pl.orders._client._base_url = TEST_URL
    filename = pl.orders.download_asset(dl_url, directory=str(tmpdir))

    with open(filename) as f:
        assert json.load(f) == {'key': 'value'}

    assert Path(filename).name == 'metadata.json'


@respx.mock
@pytest.mark.anyio
async def test_download_asset_img(tmpdir, open_test_img, session):
    dl_url = TEST_DOWNLOAD_URL + '/1?token=IAmAToken'

    img_headers = {
        'Content-Type': 'image/tiff',
        'Content-Length': '527',
        'Content-Disposition': 'attachment; filename="img.tif"'
    }

    async def _stream_img():
        data = open_test_img.read()
        v = memoryview(data)

        chunksize = 100
        for i in range(math.ceil(len(v) / (chunksize))):
            yield v[i * chunksize:min((i + 1) * chunksize, len(v))]

    # populate request parameter to avoid respx cloning, which throws
    # an error caused by respx and not this code
    # https://github.com/lundberg/respx/issues/130
    mock_resp = httpx.Response(HTTPStatus.OK,
                               stream=_stream_img(),
                               headers=img_headers,
                               request='donotcloneme')
    respx.get(dl_url).return_value = mock_resp

    cl = OrdersClient(session, base_url=TEST_URL)
    filename = await cl.download_asset(dl_url, directory=str(tmpdir))

    assert Path(filename).name == 'img.tif'
    assert os.path.isfile(filename)


@respx.mock
@pytest.mark.anyio
@pytest.mark.parametrize("checksum", [("MD5"), ("SHA256")])
@pytest.mark.parametrize(
    "asset1_bytes, expectation",
    [(b"1", does_not_raise()), (b"1", does_not_raise()),
     (b"does not match", pytest.raises(exceptions.ClientError))])
async def test_validate_checksum_checksum(tmpdir,
                                          asset1_bytes,
                                          expectation,
                                          checksum):

    itemtype1_dir = Path(tmpdir, 'itemtype1')
    itemtype1_dir.mkdir()

    asset1 = itemtype1_dir / 'asset1.tif'
    asset1.write_bytes(b"1")

    asset2 = itemtype1_dir / 'asset2.json'
    asset2.write_bytes(b'{"foo": "bar"}')
    asset2_bytes = asset2.read_bytes()

    manifest_data = {
        "name": "",
        "files": [
            {
                "path": "itemtype1/asset1.tif",
                "digests": {
                    "md5": hashlib.md5(asset1_bytes).hexdigest(),
                    "sha256": hashlib.sha256(asset1_bytes).hexdigest()}
            }, {
                "path": "itemtype1/asset2.json",
                "digests": {
                    "md5": hashlib.md5(asset2_bytes).hexdigest(),
                    "sha256": hashlib.sha256(asset2_bytes).hexdigest()}
            }]
    }  # yapf: disable
    Path(tmpdir, 'manifest.json').write_text(json.dumps(manifest_data))

    with expectation:
        OrdersClient.validate_checksum(Path(tmpdir), checksum)


@respx.mock
@pytest.mark.parametrize("checksum", [("MD5"), ("SHA256")])
@pytest.mark.parametrize(
    "asset1_bytes, expectation",
    [(b"1", does_not_raise()), (b"1", does_not_raise()),
     (b"does not match", pytest.raises(exceptions.ClientError))])
def test_validate_checksum_checksum_sync(tmpdir,
                                         asset1_bytes,
                                         expectation,
                                         checksum):

    itemtype1_dir = Path(tmpdir, 'itemtype1')
    itemtype1_dir.mkdir()

    asset1 = itemtype1_dir / 'asset1.tif'
    asset1.write_bytes(b"1")

    asset2 = itemtype1_dir / 'asset2.json'
    asset2.write_bytes(b'{"foo": "bar"}')
    asset2_bytes = asset2.read_bytes()

    manifest_data = {
        "name": "",
        "files": [
            {
                "path": "itemtype1/asset1.tif",
                "digests": {
                    "md5": hashlib.md5(asset1_bytes).hexdigest(),
                    "sha256": hashlib.sha256(asset1_bytes).hexdigest()}
            }, {
                "path": "itemtype1/asset2.json",
                "digests": {
                    "md5": hashlib.md5(asset2_bytes).hexdigest(),
                    "sha256": hashlib.sha256(asset2_bytes).hexdigest()}
            }]
    }  # yapf: disable
    Path(tmpdir, 'manifest.json').write_text(json.dumps(manifest_data))

    pl = Planet()

    with expectation:
        pl.orders.validate_checksum(Path(tmpdir), checksum)


@respx.mock
@pytest.mark.anyio
@pytest.mark.parametrize(
    "create, corrupt, expectation",
    [(True, False, does_not_raise()),
     (True, True, pytest.raises(exceptions.ClientError)),
     (False, False, pytest.raises(exceptions.ClientError))])
async def test_validate_checksum_manifest(
    tmpdir,
    create,
    corrupt,
    expectation,
):
    itemtype1_dir = Path(tmpdir, 'itemtype1')
    itemtype1_dir.mkdir()

    asset1 = itemtype1_dir / 'asset1.tif'
    asset1.write_bytes(b"1")

    Path(tmpdir, 'asset1.tif').write_bytes(b"1")

    manifest_data = {
        "name": "",
        "files": [{
            "path": "itemtype1/asset1.tif",
            "digests": {
                "md5": hashlib.md5(b"1").hexdigest(),
                "sha256": hashlib.sha256(b"1").hexdigest()
            },
        }]
    }
    if create:
        if corrupt:
            Path(tmpdir, 'manifest.json').write_text('not json')
        else:
            Path(tmpdir, 'manifest.json').write_text(json.dumps(manifest_data))

    with expectation:
        OrdersClient.validate_checksum(Path(tmpdir), 'md5')


@respx.mock
@pytest.mark.anyio
@pytest.mark.parametrize(
    "results, paths",
    [(None, []),
     ([], []),
     ([{"location": f'{TEST_DOWNLOAD_URL}/1',
        "name": "oid/itemtype1/asset.json"},
       {"location": f'{TEST_DOWNLOAD_URL}/2',
        "name": "oid/itemtype2/asset.json"},
       ],
      [Path('oid', 'itemtype1', 'asset.json'),
       Path('oid', 'itemtype2', 'asset.json'),
       ])
     ])  # yapf: disable
async def test_download_order_success(results,
                                      paths,
                                      tmpdir,
                                      order_description,
                                      oid,
                                      session):

    # Mock an HTTP response for download
    order_description['state'] = 'success'
    order_description['_links']['results'] = results

    get_url = f'{TEST_ORDERS_URL}/{oid}'
    mock_resp = httpx.Response(HTTPStatus.OK, json=order_description)
    respx.get(get_url).return_value = mock_resp

    mock_resp1 = httpx.Response(
        HTTPStatus.OK,
        json={'key': 'value'},
        headers={
            'Content-Type': 'application/json',
            'Content-Disposition': 'attachment; filename="asset.json"'
        })
    respx.get(f'{TEST_DOWNLOAD_URL}/1').return_value = mock_resp1

    mock_resp2 = httpx.Response(
        HTTPStatus.OK,
        json={'key2': 'value2'},
        headers={
            'Content-Type': 'application/json',
            'Content-Disposition': 'attachment; filename="asset.json"'
        })
    respx.get(f'{TEST_DOWNLOAD_URL}/2').return_value = mock_resp2

    cl = OrdersClient(session, base_url=TEST_URL)
    filenames = await cl.download_order(oid, directory=str(tmpdir))

    assert filenames == [Path(tmpdir, p) for p in paths]

    if filenames:
        with open(filenames[0]) as f:
            assert json.load(f) == {'key': 'value'}

        with open(filenames[1]) as f:
            assert json.load(f) == {'key2': 'value2'}


@respx.mock
@pytest.mark.parametrize(
    "results, paths",
    [(None, []),
     ([], []),
     ([{"location": f'{TEST_DOWNLOAD_URL}/1',
        "name": "oid/itemtype1/asset.json"},
       {"location": f'{TEST_DOWNLOAD_URL}/2',
        "name": "oid/itemtype2/asset.json"},
       ],
      [Path('oid', 'itemtype1', 'asset.json'),
       Path('oid', 'itemtype2', 'asset.json'),
       ])
     ])  # yapf: disable
def test_download_order_success_sync(results,
                                     paths,
                                     tmpdir,
                                     order_description,
                                     oid,
                                     session):

    # Mock an HTTP response for download
    order_description['state'] = 'success'
    order_description['_links']['results'] = results

    get_url = f'{TEST_ORDERS_URL}/{oid}'
    mock_resp = httpx.Response(HTTPStatus.OK, json=order_description)
    respx.get(get_url).return_value = mock_resp

    mock_resp1 = httpx.Response(
        HTTPStatus.OK,
        json={'key': 'value'},
        headers={
            'Content-Type': 'application/json',
            'Content-Disposition': 'attachment; filename="asset.json"'
        })
    respx.get(f'{TEST_DOWNLOAD_URL}/1').return_value = mock_resp1

    mock_resp2 = httpx.Response(
        HTTPStatus.OK,
        json={'key2': 'value2'},
        headers={
            'Content-Type': 'application/json',
            'Content-Disposition': 'attachment; filename="asset.json"'
        })
    respx.get(f'{TEST_DOWNLOAD_URL}/2').return_value = mock_resp2

    pl = Planet()
    pl.orders._client._base_url = TEST_URL
    filenames = pl.orders.download_order(oid, directory=str(tmpdir))

    assert filenames == [Path(tmpdir, p) for p in paths]

    if filenames:
        with open(filenames[0]) as f:
            assert json.load(f) == {'key': 'value'}

        with open(filenames[1]) as f:
            assert json.load(f) == {'key2': 'value2'}


@respx.mock
@pytest.mark.anyio
async def test_download_order_state(tmpdir, order_description, oid, session):
    dl_url1 = TEST_DOWNLOAD_URL + '/1?token=IAmAToken'
    order_description['_links']['results'] = [
        {
            'location': dl_url1
        },
    ]

    get_url = f'{TEST_ORDERS_URL}/{oid}'
    mock_resp = httpx.Response(HTTPStatus.OK, json=order_description)
    respx.get(get_url).return_value = mock_resp

    cl = OrdersClient(session, base_url=TEST_URL)
    with pytest.raises(exceptions.ClientError):
        await cl.download_order(oid, directory=str(tmpdir))


@respx.mock
def test_download_order_state_sync(tmpdir, order_description, oid, session):
    dl_url1 = TEST_DOWNLOAD_URL + '/1?token=IAmAToken'
    order_description['_links']['results'] = [
        {
            'location': dl_url1
        },
    ]

    get_url = f'{TEST_ORDERS_URL}/{oid}'
    mock_resp = httpx.Response(HTTPStatus.OK, json=order_description)
    respx.get(get_url).return_value = mock_resp

    pl = Planet()
    pl.orders._client._base_url = TEST_URL
    with pytest.raises(exceptions.ClientError):
        pl.orders.download_order(oid, directory=str(tmpdir))


@respx.mock
@pytest.mark.anyio
async def test_download_order_overwrite_true_preexisting_data(
        tmpdir,
        oid,
        session,
        create_download_mock,
        original_content,
        downloaded_content):
    """
    Test if download_order() overwrites pre-existing data with
    overwrite flag set to True.
    """

    # Save JSON to a temporary file
    with open(Path(tmpdir, 'file.json'), "a") as out_file:
        json.dump(original_content, out_file)

    create_download_mock()
    cl = OrdersClient(session, base_url=TEST_URL)
    await cl.download_order(oid, directory=str(tmpdir), overwrite=True)

    # Check that the data downloaded has overwritten the original data
    with open(Path(tmpdir, 'file.json')) as f:
        assert json.load(f) == downloaded_content


@respx.mock
@pytest.mark.anyio
async def test_download_order_overwrite_false_preexisting_data(
        tmpdir, oid, session, create_download_mock, original_content):
    """
    Test if download_order() does not overwrite pre-existing
    data with overwrite flag set to False.
    """

    # Save JSON to a temporary file
    with open(Path(tmpdir, 'file.json'), "a") as out_file:
        json.dump(original_content, out_file)

    create_download_mock()
    cl = OrdersClient(session, base_url=TEST_URL)
    await cl.download_order(oid, directory=str(tmpdir), overwrite=False)

    # Check that the original data has not been overwritten
    with open(Path(tmpdir, 'file.json')) as f:
        assert json.load(f) == original_content


@respx.mock
@pytest.mark.anyio
async def test_download_order_overwrite_true_nonexisting_data(
        tmpdir, oid, session, create_download_mock, downloaded_content):
    """
    Test if download_order() downloads data with overwrite flag
    set to true without pre-existing data.
    """

    create_download_mock()
    cl = OrdersClient(session, base_url=TEST_URL)
    await cl.download_order(oid, directory=str(tmpdir), overwrite=True)

    with open(Path(tmpdir, 'file.json')) as f:
        assert json.load(f) == downloaded_content


@respx.mock
@pytest.mark.anyio
async def test_download_order_overwrite_false_nonexisting_data(
        tmpdir, oid, session, create_download_mock, downloaded_content):
    """
    Test if download_order() downloads data with overwrite flag
    set to false without pre-existing data.
    """

    create_download_mock()
    cl = OrdersClient(session, base_url=TEST_URL)
    await cl.download_order(oid, directory=str(tmpdir), overwrite=False)

    # Check that the was data downloaded and has the correct contents
    with open(Path(tmpdir, 'file.json')) as f:
        assert json.load(f) == downloaded_content



from typing import Optional

from .features import FeaturesAPI
from .data import DataAPI
from .orders import OrdersAPI
from .subscriptions import SubscriptionsAPI
from planet.http import Session
from planet.__version__ import __version__

SYNC_CLIENT_X_PLANET_APP = "python-sdk-sync"


class Planet:
    """Planet API client. This client contains non-async methods.

    Authentication is required: defaults to detecting API key from environment (PL_API_KEY).

    Members:
    `data`: for interacting with the Planet Data API.
    `orders`: Orders API.
    `subscriptions`: Subscriptions API.
    `features`: Features API

    Quick start example:
    ```python
    # requires PL_API_KEY

    pl = Planet()
    for item in pl.data.search(['PSScene'], limit=5):
        print(item)

    for sub in pl.subscriptions.list_subscriptions():
        print(item)
    ```

    Parameters:
        session: Optional Session. The Session can be provided allowing for customization, and
            will default to standard behavior when not provided. Example:

          ```python
          from planet.sync import Planet

          pl = Planet()
          ````
    """

    def __init__(self, session: Optional[Session] = None) -> None:
        self._session = session or Session()
        self._session._client.headers.update({
            "X-Planet-App": SYNC_CLIENT_X_PLANET_APP,
            "User-Agent": f"planet-client-python/{__version__}/sync",
        })

        self.data = DataAPI(self._session)
        self.orders = OrdersAPI(self._session)
        self.subscriptions = SubscriptionsAPI(self._session)
        self.features = FeaturesAPI(self._session)



# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
from http import HTTPStatus
import json
import os

from click.testing import CliRunner
import httpx
import jwt
import pytest
import respx

from planet.cli import cli

TEST_URL = 'http://MockNotRealURL/api/path'
TEST_LOGIN_URL = f'{TEST_URL}/login'


# skip the global mock of _SecretFile.read
# for this module
@pytest.fixture(autouse=True, scope='module')
def test_secretfile_read():
    return


@pytest.fixture
def redirect_secretfile(tmp_path):
    """patch the cli so it works with a temporary secretfile

    this is to avoid collisions with the actual planet secretfile
    """
    secretfile_path = tmp_path / 'secret.json'

    with pytest.MonkeyPatch.context() as mp:
        mp.setattr(cli.auth.planet.auth, 'SECRET_FILE_PATH', secretfile_path)
        yield secretfile_path


@respx.mock
def test_cli_auth_init_success(redirect_secretfile):
    """Test the successful auth init path

    Also tests the base-url command, since we will get an exception
    if the base url is not changed to the mocked url
    """
    payload = {'api_key': 'test_cli_auth_init_success_key'}
    resp = {'token': jwt.encode(payload, 'key')}
    mock_resp = httpx.Response(HTTPStatus.OK, json=resp)
    respx.post(TEST_LOGIN_URL).return_value = mock_resp

    result = CliRunner().invoke(cli.main,
                                args=['auth', '--base-url', TEST_URL, 'init'],
                                input='email\npw\n')

    # we would get a 'url not mocked' exception if the base url wasn't
    # changed to the mocked url
    assert not result.exception

    assert 'Initialized' in result.output


@respx.mock
def test_cli_auth_init_bad_pw(redirect_secretfile):
    resp = {
        "errors": None,
        "message": "Invalid email or password",
        "status": 401,
        "success": False
    }
    mock_resp = httpx.Response(401, json=resp)
    respx.post(TEST_LOGIN_URL).return_value = mock_resp

    result = CliRunner().invoke(cli.main,
                                args=['auth', '--base-url', TEST_URL, 'init'],
                                input='email\npw\n')

    assert result.exception
    assert 'Error: Incorrect email or password.\n' in result.output


def test_cli_auth_value_success(redirect_secretfile):
    key = 'test_cli_auth_value_success_key'
    content = {'key': key}
    with open(redirect_secretfile, 'w') as f:
        json.dump(content, f)

    result = CliRunner().invoke(cli.main, ['auth', 'value'])
    assert not result.exception
    assert result.output == f'{key}\n'


def test_cli_auth_value_failure(redirect_secretfile):
    result = CliRunner().invoke(cli.main, ['auth', 'value'])
    assert result.exception
    assert 'Error: Auth information does not exist or is corrupted.' \
        in result.output


def test_cli_auth_store_cancel(redirect_secretfile):
    result = CliRunner().invoke(cli.main, ['auth', 'store', 'setval'],
                                input='')
    assert not result.exception
    assert not os.path.isfile(redirect_secretfile)


def test_cli_auth_store_confirm(redirect_secretfile):
    result = CliRunner().invoke(cli.main, ['auth', 'store', 'setval'],
                                input='y')
    assert not result.exception

    with open(redirect_secretfile, 'r') as f:
        assert json.load(f) == {'key': 'setval'}



# Copyright 2020 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Manage authentication with Planet APIs"""
from __future__ import annotations  # https://stackoverflow.com/a/33533514
import abc
import json
import logging
import os
import pathlib
import stat
import typing
from typing import Optional

import httpx
import jwt

from . import http
from .constants import ENV_API_KEY, PLANET_BASE_URL, SECRET_FILE_PATH
from .exceptions import AuthException

LOGGER = logging.getLogger(__name__)

BASE_URL = f'{PLANET_BASE_URL}/v0/auth'

AuthType = httpx.Auth


class Auth(metaclass=abc.ABCMeta):
    """Handle authentication information for use with Planet APIs."""

    @staticmethod
    def from_key(key: str) -> AuthType:
        """Obtain authentication from api key.

        Parameters:
            key: Planet API key
        """
        auth = APIKeyAuth(key=key)
        LOGGER.debug('Auth obtained from api key.')
        return auth

    @staticmethod
    def from_file(
        filename: Optional[typing.Union[str,
                                        pathlib.Path]] = None) -> AuthType:
        """Create authentication from secret file.

        The secret file is named `.planet.json` and is stored in the user
        directory. The file has a special format and should have been created
        with `Auth.write()`.

        Parameters:
            filename: Alternate path for the planet secret file.

        """
        filename = filename or SECRET_FILE_PATH

        try:
            secrets = _SecretFile(filename).read()
            auth = APIKeyAuth.from_dict(secrets)
        except FileNotFoundError:
            raise AuthException(f'File {filename} does not exist.')
        except (KeyError, json.decoder.JSONDecodeError):
            raise AuthException(f'File {filename} is not the correct format.')

        LOGGER.debug(f'Auth read from secret file {filename}.')
        return auth

    @staticmethod
    def from_env(variable_name: Optional[str] = None) -> AuthType:
        """Create authentication from environment variable.

        Reads the `PL_API_KEY` environment variable

        Parameters:
            variable_name: Alternate environment variable.
        """
        variable_name = variable_name or ENV_API_KEY
        api_key = os.getenv(variable_name, '')
        try:
            auth = APIKeyAuth(api_key)
            LOGGER.debug(f'Auth set from environment variable {variable_name}')
        except APIKeyAuthException:
            raise AuthException(
                f'Environment variable {variable_name} either does not exist '
                'or is empty.')
        return auth

    @staticmethod
    def from_login(email: str,
                   password: str,
                   base_url: Optional[str] = None) -> AuthType:
        """Create authentication from login email and password.

        Note: To keep your password secure, the use of `getpass` is
        recommended.

        Parameters:
            email: Planet account email address.
            password: Planet account password.
            base_url: The base URL to use. Defaults to production
                authentication API base url.
        """
        cl = AuthClient(base_url=base_url)
        auth_data = cl.login(email, password)

        api_key = auth_data['api_key']
        auth = APIKeyAuth(api_key)
        LOGGER.debug('Auth set from login email and password')
        return auth

    @classmethod
    @abc.abstractmethod
    def from_dict(cls, data: dict) -> AuthType:
        pass

    @property
    @abc.abstractmethod
    def value(self):
        pass

    @abc.abstractmethod
    def to_dict(self) -> dict:
        pass

    def store(self,
              filename: Optional[typing.Union[str, pathlib.Path]] = None):
        """Store authentication information in secret file.

        Parameters:
            filename: Alternate path for the planet secret file.
        """
        filename = filename or SECRET_FILE_PATH
        secret_file = _SecretFile(filename)
        secret_file.write(self.to_dict())


class AuthClient:

    def __init__(self, base_url: Optional[str] = None):
        """
        Parameters:
            base_url: The base URL to use. Defaults to production
                authentication API base url.
        """
        self._base_url = base_url or BASE_URL
        if self._base_url.endswith('/'):
            self._base_url = self._base_url[:-1]

    def login(self, email: str, password: str) -> dict:
        """Login using email identity and credentials.

        Note: To keep your password secure, the use of `getpass` is
        recommended.

        Parameters:
            email: Planet account email address.
            password:  Planet account password.

        Returns:
             A JSON object containing an `api_key` property with the user's
        API_KEY.
        """
        url = f'{self._base_url}/login'
        data = {'email': email, 'password': password}

        sess = http.AuthSession()
        resp = sess.request(url=url, method='POST', json=data)
        return self.decode_response(resp)

    @staticmethod
    def decode_response(response):
        """Decode the token JWT"""
        token = response.json()['token']
        return jwt.decode(token, options={'verify_signature': False})


class APIKeyAuthException(AuthException):
    """exceptions thrown by APIKeyAuth"""
    pass


class APIKeyAuth(httpx.BasicAuth, Auth):
    """Planet API Key authentication."""
    DICT_KEY = 'key'

    def __init__(self, key: str):
        """Initialize APIKeyAuth.

        Parameters:
            key: API key.

        Raises:
            APIKeyException: If API key is None or empty string.
        """
        if not key:
            raise APIKeyAuthException('API key cannot be empty.')
        self._key = key
        super().__init__(self._key, '')

    @classmethod
    def from_dict(cls, data: dict) -> APIKeyAuth:
        """Instantiate APIKeyAuth from a dict."""
        api_key = data[cls.DICT_KEY]
        return cls(api_key)

    def to_dict(self):
        """Represent APIKeyAuth as a dict."""
        return {self.DICT_KEY: self._key}

    @property
    def value(self):
        return self._key


class _SecretFile:

    def __init__(self, path: typing.Union[str, pathlib.Path]):
        self.path = pathlib.Path(path)

        self.permissions = stat.S_IRUSR | stat.S_IWUSR  # user rw

        # in sdk versions <=2.0.0, secret file was created with the wrong
        # permissions, fix this automatically as well as catching the unlikely
        # cases where the permissions get changed externally
        self._enforce_permissions()

    def write(self, contents: dict):
        try:
            secrets_to_write = self.read()
            secrets_to_write.update(contents)
        except (FileNotFoundError, KeyError, json.decoder.JSONDecodeError):
            secrets_to_write = contents

        self._write(secrets_to_write)

    def _write(self, contents: dict):
        LOGGER.debug(f'Writing to {self.path}')

        def opener(path, flags):
            return os.open(path, flags, self.permissions)

        with open(self.path, 'w', opener=opener) as fp:
            fp.write(json.dumps(contents))

    def read(self) -> dict:
        LOGGER.debug(f'Reading from {self.path}')
        with open(self.path, 'r') as fp:
            contents = json.loads(fp.read())
        return contents

    def _enforce_permissions(self):
        """if the file's permissions are not what they should be, fix them"""
        if self.path.exists():
            # in octal, permissions is the last three bits of the mode
            file_permissions = self.path.stat().st_mode & 0o777
            if file_permissions != self.permissions:
                LOGGER.info('Fixing planet secret file permissions.')
                self.path.chmod(self.permissions)



# Copyright 2023 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Functionality for preparing subscription requests."""
from datetime import datetime
from dataclasses import dataclass, asdict
from typing import Any, Dict, Optional, List, Mapping, Sequence, Union

from typing_extensions import Literal

from . import geojson, specs
from .exceptions import ClientError

NOTIFICATIONS_TOPICS = ('delivery.success',
                        'delivery.match',
                        'delivery.failed',
                        'status.backfill.completed',
                        'status.completed',
                        'status.cancelled',
                        'status.pending',
                        'status.all',
                        'status.suspended',
                        'status.failed')

REPROJECT_KERNEL = ('near',
                    'bilinear',
                    'cubic',
                    'cubicspline',
                    'lanczos',
                    'average',
                    'mode',
                    'min',
                    'max',
                    'med',
                    'q1',
                    'q3')
REPROJECT_KERNEL_DEFAULT = 'near'


def build_request(name: str,
                  source: Mapping,
                  delivery: Optional[Mapping] = None,
                  notifications: Optional[Mapping] = None,
                  tools: Optional[List[Mapping]] = None,
                  hosting: Optional[Union[Mapping, str]] = None,
                  collection_id: Optional[str] = None,
                  create_configuration: Optional[bool] = False,
                  clip_to_source: Optional[bool] = False) -> dict:
    """Construct a Subscriptions API request.

    The return value can be passed to
    [planet.clients.subscriptions.SubscriptionsClient.create_subscription][].

    Parameters:
        name: Name of the subscription.
        source: A source for the subscription, i.e. catalog.
        delivery: A delivery mechanism e.g. GCS, AWS, Azure, or OCS.
        notifications: Specify notifications via email/webhook.
        tools: Tools to apply to the products. The order of operation
            is determined by the service.
        hosting: A hosting destination e.g. Sentinel Hub.
        collection_id: A Sentinel Hub collection ID.
        create_configuration: Automatically create a layer configuration for your collection.
        clip_to_source: whether to clip to the source geometry or not
            (the default). If True a clip configuration will be added to
            the list of requested tools unless an existing clip tool
            exists.  NOTE: Not all data layers support clipping, please
            consult the Product reference before using this option.
            NOTE: the next version of the Subscription API will remove
            the clip tool option and always clip to the source geometry.
            Thus this is a preview of the next API version's default
            behavior.

    Returns:
        dict: a representation of a Subscriptions API request for
            a new subscription.

    Raises:
        ClientError: when a valid Subscriptions API request can't be
            constructed.

    Example::

        from datetime import datetime
        from planet.subscription_request import build_request, catalog_source, amazon_s3

        geom = {
            "coordinates": [
                [
                    [139.5648193359375, 35.42374884923695],
                    [140.1031494140625, 35.42374884923695],
                    [140.1031494140625, 35.77102915686019],
                    [139.5648193359375, 35.77102915686019],
                    [139.5648193359375, 35.42374884923695],
                ]
            ],
            "type": "Polygon",
        }

        source = catalog_source(["PSScene"], ["ortho_analytic_4b"], geom, datetime(2021, 3, 1))

        delivery = amazon_s3(ACCESS_KEY_ID, SECRET_ACCESS_KEY, "test", "us-east-1")

        hosting = sentinel_hub("2716077c-191e-4e47-9e3f-01c9c429f88d")

        subscription_request = build_request(
            "test_subscription", source=source, delivery=delivery, hosting=hosting
        )
    """
    # Because source is a Mapping we must make copies for
    # the function's return value. dict() shallow copies a Mapping
    # and returns a new dict.
    details = {"name": name, "source": dict(source)}

    if delivery:
        details['delivery'] = dict(delivery)

    if notifications:
        details['notifications'] = dict(notifications)

    if tools or clip_to_source:
        tool_list = [dict(tool) for tool in (tools or [])]

        # If clip_to_source is True a clip configuration will be added
        # to the list of requested tools unless an existing clip tool
        # exists. In that case an exception is raised. NOTE: the next
        # version of the Subscription API will remove the clip tool
        # option and always clip to the source geometry. Thus this is a
        # preview of the next API version's default behavior.
        if clip_to_source:
            if any(tool.get('type', None) == 'clip' for tool in tool_list):
                raise ClientError(
                    "clip_to_source option conflicts with a configured clip tool."
                )
            else:
                tool_list.append({
                    'type': 'clip',
                    'parameters': {
                        'aoi': source['parameters']['geometry']
                    }
                })

        details['tools'] = tool_list

    if hosting == "sentinel_hub":
        hosting_info: Dict[str, Any] = {
            "type": "sentinel_hub", "parameters": {}
        }
        if collection_id:
            hosting_info["parameters"]["collection_id"] = collection_id
        if create_configuration:
            hosting_info["parameters"][
                "create_configuration"] = create_configuration
        details['hosting'] = hosting_info
    elif isinstance(hosting, dict):
        details['hosting'] = hosting

    return details


def catalog_source(
    item_types: List[str],
    asset_types: List[str],
    geometry: Union[dict, str],
    start_time: datetime,
    filter: Optional[Mapping] = None,
    end_time: Optional[datetime] = None,
    rrule: Optional[str] = None,
    publishing_stages: Optional[Sequence[Literal["preview",
                                                 "standard",
                                                 "finalized"]]] = None,
    time_range_type: Optional[Literal["acquired", "published"]] = None,
) -> dict:
    """Construct a Catalog subscription source.

    The return value can be passed to
    [planet.subscription_request.build_request][].

    Parameters:
        item_types: The class of spacecraft and processing level of the
            subscription's matching items, e.g. PSScene.
        asset_types: The data products which will be delivered for all
            subscription matching items. An item will only match and
            deliver if all specified asset types are published for that
            item.
        geometry: The area of interest of the subscription that will be
            used to determine matches. Can be either a geojson-like dict
            or a Features API feature reference (string)
        start_time: The start time of the subscription. This time can be
            in the past or future.
        filter: The filter criteria based on item-level metadata.
        end_time: The end time of the subscription. This time can be in
            the past or future, and must be after the start_time.
        rrule: The recurrence rule, given in iCalendar RFC 5545 format.
            Only monthly recurrences are supported at this time.
        publishing_stages: A sequence of one or more of the values
            "preview", "standard", or "finalized".
        time_range_type: "acquired" (new in 2.1.0) or "published".

    Returns:
        dict: a representation of a subscription source.

    Raises:
        ClientError: if a source can not be
            configured.

    Examples:

        ```python
        source = catalog_source(
            ["PSScene"],
            ["ortho_analytic_4b"],
            geometry={
                "type": "Polygon",
                "coordinates": [
                    [
                        [37.791595458984375, 14.84923123791421],
                        [37.90214538574219, 14.84923123791421],
                        [37.90214538574219, 14.945448293647944],
                        [37.791595458984375, 14.945448293647944],
                        [37.791595458984375, 14.84923123791421],
                    ]
                ],
            },
            start_time=datetime(2021, 3, 1),
            publishing_stages=["standard"],
            time_range_type="acquired",
        )

        request = build_request("Standard PSScene Ortho Analytic", source=source, delivery={})
        ```
    """
    if len(item_types) > 1:
        raise ClientError(
            "Subscription can only be successfully created if one item type",
            "is specified.")
    try:
        asset_types = [
            specs.validate_asset_type(item, asset) for asset in asset_types
            for item in item_types
        ]
    except specs.SpecificationException as exception:
        raise ClientError(exception)

    parameters = {
        "item_types": item_types,
        "asset_types": asset_types,
        "geometry": geojson.as_geom_or_ref(geometry),
    }

    try:
        parameters['start_time'] = _datetime_to_rfc3339(start_time)
    except AttributeError:
        raise ClientError('Could not convert start_time to an iso string')

    if filter:
        parameters['filter'] = dict(filter)

    if end_time:
        try:
            parameters['end_time'] = _datetime_to_rfc3339(end_time)
        except AttributeError:
            raise ClientError('Could not convert end_time to an iso string')

    if rrule:
        parameters['rrule'] = rrule

    if publishing_stages:
        parameters['publishing_stages'] = list(set(publishing_stages))

    if time_range_type:
        parameters['time_range_type'] = time_range_type

    return {"parameters": parameters}


def planetary_variable_source(
    var_type: Optional[str],
    var_id: str,
    geometry: Union[dict, str],
    start_time: datetime,
    end_time: Optional[datetime] = None,
) -> dict:
    """Construct a Planetary Variable subscription source.

    Planetary Variables come in 4 types and are further subdivided
    within these types. See [Subscribing to Planetary Variables](https://developers.planet.com/docs/subscriptions/pvs-subs/#planetary-variables-types-and-ids)
    or the [OpenAPI spec](https://api.planet.com/subscriptions/v1/spec) for
    more details.

    The return value can be passed to
    [planet.subscription_request.build_request][].

    Note: this function does not validate variable types and ids.

    Parameters:
        var_type: Planetary Variable type. See documentation for all
            available types.  Used to be a required parameter but
            is now optional and can be 'None'.
        var_id: A Planetary Variable ID. See documenation for all
            available IDs.
        geometry: The area of interest of the subscription that will be
            used to determine matches. May be a geojson-like dict or a
            Features API geometry reference (string)
        start_time: The start time of the subscription. This time can be
            in the past or future.
        end_time: The end time of the subscription. This time can be in
            the past or future, and must be after the start_time.

    Returns:
        dict: a representation of a subscription source.

    Raises:
        ClientError: if a source can not be
            configured.

    Examples:

        ```python
        pv_source = planetary_variables_source(
            "soil_water_content",
            "SWC-AMSR2-C_V1.0_100",
            geometry={
                "type": "Polygon",
                "coordinates": [
                    [
                        [37.791595458984375, 14.84923123791421],
                        [37.90214538574219, 14.84923123791421],
                        [37.90214538574219, 14.945448293647944],
                        [37.791595458984375, 14.945448293647944],
                        [37.791595458984375, 14.84923123791421],
                    ]
                ],
            },
            start_time=datetime(2021, 3, 1),
        )

        request = build_request(
            "soil_water_subscription",
            source=pv_source,
            delivery={},
        )
        ```
    """
    # TODO: validation of variable types and ids.

    parameters = {
        "id": var_id,
        "geometry": geojson.as_geom_or_ref(geometry),
    }

    try:
        parameters['start_time'] = _datetime_to_rfc3339(start_time)
    except AttributeError:
        raise ClientError('Could not convert start_time to an iso string')

    if end_time:
        try:
            parameters['end_time'] = _datetime_to_rfc3339(end_time)
        except AttributeError:
            raise ClientError('Could not convert end_time to an iso string')

    source: dict[str, Any] = {"parameters": parameters}
    if var_type:
        source["type"] = var_type
    return source


def _datetime_to_rfc3339(value: datetime) -> str:
    """Converts the datetime to an RFC3339 string"""
    iso = value.isoformat()
    if value.utcoffset() is None:
        # rfc3339 needs a Z if there is no timezone offset
        iso += 'Z'
    return iso


def _delivery(type: str, parameters: dict) -> dict:
    return {"type": type, "parameters": parameters}


def amazon_s3(aws_access_key_id: str,
              aws_secret_access_key: str,
              bucket: str,
              aws_region: str,
              path_prefix: Optional[str] = None) -> dict:
    """Delivery to Amazon S3.

    Parameters:
        aws_access_key_id: S3 account access key.
        aws_secret_access_key: S3 account secret key.
        bucket: The name of the bucket that will receive the order output.
        aws_region: The region where the bucket lives in AWS.
        path_prefix: Path prefix for deliveries.
    """
    parameters = {
        'aws_access_key_id': aws_access_key_id,
        'aws_secret_access_key': aws_secret_access_key,
        'bucket': bucket,
        'aws_region': aws_region,
    }

    if path_prefix:
        parameters['path_prefix'] = path_prefix

    return _delivery('amazon_s3', parameters)


def azure_blob_storage(account: str,
                       container: str,
                       sas_token: str,
                       storage_endpoint_suffix: Optional[str] = None,
                       path_prefix: Optional[str] = None) -> dict:
    """Delivery to Azure Blob Storage.

    Parameters:
        account: Azure account.
        container: ABS container name.
        sas_token: Shared-Access Signature token. Token should be specified
            without a leading '?'.
        storage_endpoint_suffix: Deliver order to a sovereign cloud. The
            default is "core.windows.net".
        path_prefix: Path prefix for deliveries.
    """
    parameters = {
        'account': account,
        'container': container,
        'sas_token': sas_token,
    }

    if storage_endpoint_suffix:
        parameters['storage_endpoint_suffix'] = storage_endpoint_suffix

    if path_prefix:
        parameters['path_prefix'] = path_prefix

    return _delivery('azure_blob_storage', parameters)


def google_earth_engine(project: str, collection: str,
                        credentials: str) -> dict:
    """Delivery to Google Earth Engine.

    Parameters:
        project: GEE project name.
        collection: GEE Image Collection name.
        credentials: GEE service account credentials.
    """
    parameters = {
        'project': project,
        'collection': collection,
        'credentials': credentials
    }
    return _delivery('google_earth_engine', parameters)


def google_cloud_storage(credentials: str,
                         bucket: str,
                         path_prefix: Optional[str] = None) -> dict:
    """Delivery to Google Cloud Storage.

    Parameters:
        credentials: JSON-string of service account for bucket.
        bucket: GCS bucket name.
        path_prefix: Path prefix for deliveries.
    """
    parameters = {
        'bucket': bucket,
        'credentials': credentials,
    }

    if path_prefix:
        parameters['path_prefix'] = path_prefix

    return _delivery('google_cloud_storage', parameters)


def oracle_cloud_storage(customer_access_key_id: str,
                         customer_secret_key: str,
                         bucket: str,
                         region: str,
                         namespace: str,
                         path_prefix: Optional[str] = None) -> dict:
    """Delivery to Oracle Cloud Storage.

    Parameters:
        customer_access_key_id: Customer Secret Key credentials.
        customer_secret_key: Customer Secret Key credentials.
        bucket: The name of the bucket that will receive the order output.
        region: The region where the bucket lives in Oracle.
        namespace: Object Storage namespace name.
        path_prefix: Path prefix for deliveries.
    """
    parameters = {
        'customer_access_key_id': customer_access_key_id,
        'customer_secret_key': customer_secret_key,
        'bucket': bucket,
        'region': region,
        'namespace': namespace
    }

    if path_prefix:
        parameters['path_prefix'] = path_prefix

    return _delivery('oracle_cloud_storage', parameters)


def notifications(url: str, topics: List[str]) -> dict:
    """Specify a subscriptions API notification.

    Webhook notifications proactively notify you when a subscription matches
    and delivers an item so you have confidence that you have all the expected
    imagery.

    Parameters:
        url: location of webhook/callback where you expect to receive updates.
        topics: Event types that you can choose to be notified about.
    """
    for i, t in enumerate(topics):
        try:
            topics[i] = specs.get_match(t, NOTIFICATIONS_TOPICS, 'topic')
        except specs.SpecificationException as e:
            raise ClientError(e)

    return {"webhook": {"url": url, "topics": topics}}


def _tool(type: str, parameters: dict) -> dict:
    return {"type": type, "parameters": parameters}


def band_math_tool(b1: str,
                   b2: Optional[str] = None,
                   b3: Optional[str] = None,
                   b4: Optional[str] = None,
                   b5: Optional[str] = None,
                   b6: Optional[str] = None,
                   b7: Optional[str] = None,
                   b8: Optional[str] = None,
                   b9: Optional[str] = None,
                   b10: Optional[str] = None,
                   b11: Optional[str] = None,
                   b12: Optional[str] = None,
                   b13: Optional[str] = None,
                   b14: Optional[str] = None,
                   b15: Optional[str] = None,
                   pixel_type: str = specs.BAND_MATH_PIXEL_TYPE_DEFAULT):
    """Specify a subscriptions API band math tool.

    The parameters of the bandmath tool define how each output band in the
    derivative product should be produced, referencing the product inputs’
    original bands. Band math expressions may not reference neighboring pixels,
    as non-local operations are not supported. The tool can calculate up to 15
    bands for an item. Input band parameters may not be skipped. For example,
    if the b4 parameter is provided, then b1, b2, and b3 parameters are also
    required.

    For each band expression, the bandmath tool supports normal arithmetic
    operations and simple math operators offered in the Python numpy package.
    (For a list of supported mathematical functions, see
    [Bandmath supported numpy math routines](https://developers.planet.com/apis/orders/bandmath-numpy-routines/)).

    One bandmath imagery output file is produced for each product bundle, with
    output bands derived from the band math expressions. nodata pixels are
    processed with the band math equation. These files have “_bandmath”
    appended to their file names.

    The tool passes through UDM, RPC, and XML files, and does not update values
    in these files.

    Parameters:
        b1-b15: An expression defining how the output band should be computed.
        pixel_type: A value indicating what the output pixel type should be.

    Raises:
        planet.exceptions.ClientError: If pixel_type is not valid.
    """  # noqa
    try:
        pixel_type = specs.get_match(pixel_type,
                                     specs.BAND_MATH_PIXEL_TYPE,
                                     'pixel_type')
    except specs.SpecificationException as e:
        raise ClientError(e)

    # e.g. {"b1": "b1", "b2":"arctan(b1)"} if b1 and b2 are specified
    parameters = dict((k, v) for k, v in locals().items() if v)
    return _tool('bandmath', parameters)


def clip_tool(aoi: Mapping) -> dict:
    """Specify a subscriptions API clip tool.

    Imagery and udm files will be clipped to your area of interest. nodata
    pixels will be preserved. Xml file attributes “filename”, “numRows”,
    “numColumns” and “footprint” will be updated based on the clip results.

    The clipped output files will have “_clip” appended to their file names. If
    the clip aoi is so large that full scenes may be delivered without any
    clipping, those files will not have “_clip” appended to their file name.

    Parameters:
        aoi: GeoJSON polygon or multipolygon defining the clip area, with up to
            500 vertices. The minimum geographic area of any polygon or
            internal ring is one square meter.

    Raises:
        planet.exceptions.ClientError: If aoi is not a valid polygon or
            multipolygon.
    """
    valid_types = ['Polygon', 'MultiPolygon', 'ref']

    geom = geojson.as_geom_or_ref(dict(aoi))
    if geom['type'].lower() not in [v.lower() for v in valid_types]:
        raise ClientError(
            f'Invalid geometry type: {geom["type"]} is not in {valid_types}.')

    return _tool('clip', {'aoi': geom})


def file_format_tool(file_format: str) -> dict:
    """Specify a subscriptions API file format tool.

    Parameters:
        file_format: The format of the tool output. Either "COG" or "PL_NITF".

    Raises:
        planet.exceptions.ClientError: If file_format is not valid.
    """
    try:
        file_format = specs.validate_file_format(file_format)
    except specs.SpecificationException as e:
        raise ClientError(e)

    return _tool('file_format', {'format': file_format})


def harmonize_tool(target_sensor: str) -> dict:
    """Specify a subscriptions API harmonize tool.

    Each sensor value transforms items captured by a defined set of instrument
    IDs. Items which have not been captured by that defined set of instrument
    IDs are unaffected by (passed through) the harmonization operation.

    Parameters:
        target_sensor: A value indicating to what sensor the input asset types
            should be calibrated.

    Raises:
        planet.exceptions.ClientError: If target_sensor is not valid.
    """
    try:
        target_sensor = specs.get_match(target_sensor,
                                        specs.HARMONIZE_TOOL_TARGET_SENSORS,
                                        'target_sensor')
    except specs.SpecificationException as e:
        raise ClientError(e)

    return _tool('harmonize', {'target_sensor': target_sensor})


def reproject_tool(projection: str,
                   resolution: Optional[float] = None,
                   kernel: str = REPROJECT_KERNEL_DEFAULT) -> dict:
    """Specify a subscriptions API reproject tool.

    Parameters:
        projection: A coordinate system in the form EPSG:n (for example,
            EPSG:4326 for WGS84, EPSG:32611 for UTM 11 North (WGS84), or
            EPSG:3857 for Web Mercator). Well known text CRS values are also
            supported (for example, WGS84).
        resolution: The pixel width and height in the output file. If not
            provided, the default is the resolution of the input item. This
            value is in meters unless the coordinate system is geographic (such
            as EPSG:4326), in which case, it is pixel size in decimal degrees.
        kernel: The resampling kernel used. UDM files always use "near".

    Raises:
        planet.exceptions.ClientError: If kernel is not valid.
    """
    try:
        kernel = specs.get_match(kernel, REPROJECT_KERNEL, 'kernel')
    except specs.SpecificationException as e:
        raise ClientError(e)

    parameters: Dict[str, Any] = {"projection": projection, "kernel": kernel}
    if resolution is not None:
        parameters['resolution'] = resolution

    return _tool('reproject', parameters)


def toar_tool(scale_factor: int = 10000) -> dict:
    """Specify a subscriptions API reproject tool.

    The toar tool supports the analytic asset type for PSScene
    and REOrthoTile item types. In addition to the analytic asset, the
    corresponding XML metadata asset type is required.

    Parameters:
        scale_factor: Scale factor applied to convert 0.0 to 1.0 reflectance
            floating point values to a value that fits in 16bit integer pixels.
            The API default is 10000. Values over 65535 could result in high
            reflectances not fitting in 16bit integers.
    """
    return _tool('toar', {'scale_factor': scale_factor})


@dataclass
class FilterValue:
    """Represents a filter value with optional greater than or equal to (gte)
    and less than or equal to (lte) constraints.

    Attributes:
        gte (Optional[float]): The minimum threshold value for the filter.
        lte (Optional[float]): The maximum threshold value for the filter.
    """

    gte: Optional[float] = None
    lte: Optional[float] = None


def cloud_filter_tool(
    clear_percent: Optional[FilterValue] = None,
    cloud_percent: Optional[FilterValue] = None,
    shadow_percent: Optional[FilterValue] = None,
    heavy_haze_percent: Optional[FilterValue] = None,
    light_haze_percent: Optional[FilterValue] = None,
    snow_ice_percent: Optional[FilterValue] = None,
) -> Dict[str, Dict[str, Union[float, int]]]:
    """Specify a subscriptions API cloud_filter tool.

    The cloud_filter tool filters imagery after the clip tool has run and certain
    metadata values have been updated to pertain to the clip AOI. This tool offers
    a more detailed filtering of cloudy imagery than what can be achieved using
    only catalog source filters. For instance, you might want to receive only images
    that, after clipping, have a cloud_percent value of less than or equal to 25%.

    Parameters:
        clear_percent: Filters for images based on the percentage of clear sky.
        cloud_percent: Filters for images based on the percentage of cloud cover.
        shadow_percent: Filters for images based on the percentage of shadow cover.
        heavy_haze_percent: Filters for images based on the percentage of heavy haze cover.
        light_haze_percent: Filters for images based on the percentage of light haze cover.
        snow_ice_percent: Filters for images based on the percentage of snow or ice cover.
    """
    filters = {
        "clear_percent": clear_percent,
        "cloud_percent": cloud_percent,
        "shadow_percent": shadow_percent,
        "heavy_haze_percent": heavy_haze_percent,
        "light_haze_percent": light_haze_percent,
        "snow_ice_percent": snow_ice_percent,
    }

    result = {}

    for key, value in filters.items():
        if value:
            inner_dict = asdict(value)
            result[key] = {
                k: v
                for k, v in inner_dict.items() if v is not None
            }

    return _tool("cloud_filter", result)


def _hosting(type: str, parameters: dict) -> dict:
    return {"type": type, "parameters": parameters}


def sentinel_hub(collection_id: Optional[str],
                 create_configuration: Optional[bool] = False) -> dict:
    """Specify a Sentinel Hub hosting destination.

    Requires the user to have a Sentinel Hub account linked with their Planet
    account. Subscriptions API will create a new collection to deliver data to
    if collection_id is omitted from the request. Will also create a new layer
    configuration for the collection if create_configuration is True.
    collection_id and create_configuration are mutually exclusive in the API.

    Parameters:
        collection_id: Sentinel Hub collection
        create_configuration: Automatically create a layer configuration for your collection.
    """

    parameters: Dict[str, Union[str, bool]] = {}
    if collection_id:
        parameters['collection_id'] = collection_id
    if create_configuration:
        parameters['create_configuration'] = create_configuration
    return _hosting("sentinel_hub", parameters)



# Copyright 2020 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Functionality for preparing order details for use in creating an order"""
from __future__ import annotations  # https://stackoverflow.com/a/33533514
import logging
from typing import Any, Dict, List, Mapping, Optional, Union

from . import geojson, specs
from .exceptions import ClientError

LOGGER = logging.getLogger(__name__)


def build_request(name: str,
                  products: List[dict],
                  subscription_id: int = 0,
                  delivery: Optional[dict] = None,
                  notifications: Optional[dict] = None,
                  order_type: Optional[str] = None,
                  tools: Optional[List[dict]] = None,
                  stac: Optional[dict] = None,
                  hosting: Optional[str] = None,
                  collection_id: Optional[str] = None,
                  create_configuration: Optional[bool] = False) -> dict:
    """Prepare an order request.

    ```python
    >>> from planet.order_request import (
    ...     build_request, product, toar_tool, reproject_tool, tile_tool)
    ...
    >>> products = [
    ...     product(['20170614_113217_3163208_RapidEye-5'],
    ...             'analytic', 'REOrthoTile')])
    ... ]
    ...
    >>> tools = [
    ...     toar_tool(scale_factor=10000),
    ...     reproject_tool(projection='WSG84', kernel='cubic'),
    ...     tile_tool(1232, origin_x=-180, origin_y=-90,
    ...               pixel_size=0.000027056277056,
    ...               name_template='C1232_30_30_{tilex:04d}_{tiley:04d}')
    ... ]
    ...
    >>> order_request = build_request(
    ...     'test_order', products, tools)
    ...

    ```

    Parameters:
        name: Name of the order.
        products: Product(s) from the Data API to order.
        subscription_id: Apply this orders against this quota subscription.
        delivery: Specify custom delivery handling.
        notifications: Specify custom notifications handling.
        order_type: Accept a partial order, indicated by 'partial'.
        tools: Tools to apply to the products. Order defines
            the toolchain order of operations.
        stac: Include STAC metadata.
        hosting: A hosting destination (e.g. Sentinel Hub).
        collection_id: A Sentinel Hub collection ID.
        create_configuration: Automatically create a layer configuration for your collection.

    Raises:
        planet.specs.SpecificationException: If order_type is not a valid
            order type.
    """
    details: Dict[str, Any] = {'name': name, 'products': products}

    if subscription_id:
        details['subscription_id'] = subscription_id

    if delivery:
        details['delivery'] = delivery

    if notifications:
        details['notifications'] = notifications

    if order_type:
        validated_order_type = specs.validate_order_type(order_type)
        details['order_type'] = validated_order_type

    if tools:
        details['tools'] = tools

    if stac:
        details['metadata'] = stac

    if hosting == 'sentinel_hub':
        details['hosting'] = sentinel_hub(collection_id, create_configuration)

    return details


def product(item_ids: List[str],
            product_bundle: str,
            item_type: str,
            fallback_bundle: Optional[Union[str, List[str]]] = None) -> dict:
    """Product description for an order detail.

    Parameters:
        item_ids: IDs of the catalog items to include in the order.
        product_bundle: Set of asset types for the catalog items.
        item_type: The class of spacecraft and processing characteristics
            for the catalog items.
        fallback_bundle: In case of product_bundle not having
            all asset types available, which would result in failed
            delivery, try one or more fallback bundles. Multiple
            fallback bundles may be provided as a list or a comma
            separated string.

    Raises:
        planet.specs.SpecificationException: If bundle or fallback bundle
            are not valid bundles or if item_type is not valid for the given
            bundle or fallback bundle.
    """
    item_type = specs.validate_item_type(item_type)
    validated_product_bundle = specs.validate_bundle(item_type, product_bundle)

    if fallback_bundle is not None:
        bundles = fallback_bundle.split(',') if isinstance(
            fallback_bundle, str) else fallback_bundle
        validated_bundles = []
        for bundle in bundles:
            validated_bundles.append(specs.validate_bundle(item_type, bundle))
        validated_product_bundle = ','.join(
            [validated_product_bundle, ','.join(validated_bundles)])

    product_dict = {
        'item_ids': item_ids,
        'item_type': item_type,
        'product_bundle': validated_product_bundle
    }
    return product_dict


def notifications(email: Optional[bool] = None,
                  webhook_url: Optional[str] = None,
                  webhook_per_order: Optional[bool] = None) -> dict:
    """Notifications description for an order detail.

    Parameters:
        email: Enable email notifications for an order.
        webhook_url: URL for notification when the order is ready.
        webhook_per_order: Request a single webhook call per order instead
            of one call per each delivered item.
    """
    notifications_dict: Dict[str, Union[dict, bool]] = {}

    if webhook_url:
        webhook_dict: Dict[str, Union[str, bool]] = {
            'url': webhook_url,
        }
        if webhook_per_order is not None:
            wpo: bool = webhook_per_order
            webhook_dict['per_order'] = wpo

        notifications_dict['webhook'] = webhook_dict

    if email is not None:
        val: bool = email
        notifications_dict['email'] = val

    return notifications_dict


def delivery(archive_type: Optional[str] = None,
             single_archive: Optional[bool] = False,
             archive_filename: Optional[str] = None,
             cloud_config: Optional[Mapping] = None) -> dict:
    """Order delivery configuration.

    Example:
        ```python
        amazon_s3_config = amazon_s3(
            'access_key',
            'secret_access_key',
            'bucket_name',
            'us-east-2',
            'folder1/prefix/'
        )
        delivery_config = delivery(
            archive_type='zip',
            single_archive=True,
            archive_filename='{{order_id}}.zip'
            cloud_config=amazon_s3_config
        )
        ```

    Parameters:
        archive_type: Archive order files. Only supports 'zip'.
        single_archive: Archive all bundles together in a single file.
        archive_filename: Custom naming convention to use to name the
            archive file that is received. Uses the template variables
            {{name}} and {{order_id}}. e.g. "{{name}}_{{order_id}}.zip".
        cloud_config: Cloud delivery configuration.

    Raises:
        planet.specs.SpecificationException: If archive_type is not valid.
    """
    config: Dict[str, Any] = {}

    if archive_type:
        archive_type = specs.validate_archive_type(archive_type)

        if archive_filename is None:
            archive_filename = "{{name}}_{{order_id}}.zip"

        config.update(archive_type=archive_type,
                      archive_filename=archive_filename,
                      single_archive=single_archive)

    if cloud_config:
        config.update(cloud_config)

    return config


def amazon_s3(aws_access_key_id: str,
              aws_secret_access_key: str,
              bucket: str,
              aws_region: str,
              path_prefix: Optional[str] = None) -> dict:
    """Amazon S3 Cloud configuration.

    Parameters:
        aws_access_key_id: S3 account access key.
        aws_secret_access_key: S3 account secret key.
        bucket: The name of the bucket that will receive the order output.
        aws_region: The region where the bucket lives in AWS.
        path_prefix: Custom string to prepend to the files delivered to the
            bucket. A slash (/) character will be treated as a "folder".
            Any other characters will be added as a prefix to the files.
    """
    cloud_details = {
        'aws_access_key_id': aws_access_key_id,
        'aws_secret_access_key': aws_secret_access_key,
        'bucket': bucket,
        'aws_region': aws_region,
    }

    if path_prefix:
        cloud_details['path_prefix'] = path_prefix

    return {'amazon_s3': cloud_details}


def azure_blob_storage(account: str,
                       container: str,
                       sas_token: str,
                       storage_endpoint_suffix: Optional[str] = None,
                       path_prefix: Optional[str] = None) -> dict:
    """Azure Blob Storage configuration.

    Parameters:
        account: Azure account.
        container: ABS container name.
        sas_token: Shared-Access Signature token. Token should be specified
            without a leading '?'.
        storage_endpoint_suffix: Deliver order to a sovereign cloud.
        path_prefix: Custom string to prepend to the files delivered to the
            bucket. A slash (/) character will be treated as a "folder".
            Any other characters will be added as a prefix to the files.
    """
    cloud_details = {
        'account': account,
        'container': container,
        'sas_token': sas_token,
    }

    if storage_endpoint_suffix:
        cloud_details['storage_endpoint_suffix'] = storage_endpoint_suffix

    if path_prefix:
        cloud_details['path_prefix'] = path_prefix

    return {'azure_blob_storage': cloud_details}


def google_cloud_storage(bucket: str,
                         credentials: str,
                         path_prefix: Optional[str] = None) -> dict:
    """Google Cloud Storage configuration.

    Parameters:
        bucket: GCS bucket name.
        credentials: JSON-string of service account for bucket.
        path_prefix: Custom string to prepend to the files delivered to the
            bucket. A slash (/) character will be treated as a "folder".
            Any other characters will be added as a prefix to the files.
    """
    cloud_details = {
        'bucket': bucket,
        'credentials': credentials,
    }

    if path_prefix:
        cloud_details['path_prefix'] = path_prefix

    return {'google_cloud_storage': cloud_details}


def google_earth_engine(project: str, collection: str) -> dict:
    """Google Earth Engine configuration.

    Parameters:
        project: GEE project name.
        collection: GEE Image Collection name.
    """
    cloud_details = {
        'project': project,
        'collection': collection,
    }
    return {'google_earth_engine': cloud_details}


def oracle_cloud_storage(customer_access_key_id: str,
                         customer_secret_key: str,
                         bucket: str,
                         region: str,
                         namespace: str,
                         path_prefix: Optional[str] = None) -> dict:
    """Oracle Cloud Storage configuration.

    Parameters:
        customer_access_key_id: Customer Access Key credentials.
        customer_secret_key: Customer Secret Key credentials.
        bucket: The name of the bucket that will receive the order output.
        region: The region where the bucket lives in Oracle.
        namespace: Object Storage namespace name.
        path_prefix: Custom string to prepend to the files delivered to the
            bucket. A slash (/) character will be treated as a "folder".
            Any other characters will be added as a prefix to the files.
    """
    cloud_details = {
        'customer_access_key_id': customer_access_key_id,
        'customer_secret_key': customer_secret_key,
        'bucket': bucket,
        'region': region,
        'namespace': namespace
    }

    if path_prefix:
        cloud_details['path_prefix'] = path_prefix

    return {'oracle_cloud_storage': cloud_details}


def _tool(name: str, parameters: dict) -> dict:
    """Create the API spec representation of a tool.

    See [Tools and Toolchains](
    https://developers.planet.com/docs/orders/tools-toolchains/)
    for more information on available tools and tool parameters.

    Parameters:
        name: Tool name.
        parameters: Tool parameters.

    Raises:
        planet.specs.SpecificationException: If name is not the name of a valid
            Orders API tool.
    """
    name = specs.validate_tool(name)
    return {name: parameters}


def clip_tool(aoi: dict) -> dict:
    """Create the API spec representation of a clip tool.

    Example:
        ```python
        aoi = {
            "type": "Polygon",
            "coordinates": [[
                [37.791595458984375, 14.84923123791421],
                [37.90214538574219, 14.84923123791421],
                [37.90214538574219, 14.945448293647944],
                [37.791595458984375, 14.945448293647944],
                [37.791595458984375, 14.84923123791421]
            ]]
          }
        tool = clip_tool(aoi)
        ```

    Parameters:
        aoi: clip GeoJSON, either Polygon or Multipolygon.

    Raises:
        planet.exceptions.ClientError: If GeoJSON is not a valid polygon or
            multipolygon.
    """
    valid_types = ['Polygon', 'MultiPolygon', 'ref']

    geom = geojson.as_geom_or_ref(aoi)
    if geom['type'].lower() not in [v.lower() for v in valid_types]:
        raise ClientError(
            f'Invalid geometry type: {geom["type"]} is not in {valid_types}.')
    return _tool('clip', {'aoi': geom})


def composite_tool(group_by: Optional[str] = None) -> dict:
    """
    Create the API spec representation of a composite tool.

    Parameters:
        group_by: (Optional) Defines how items are grouped to create one or more composites.
                  Supported values are:
                  - "order": All input items are used to create a single composite output. This is the default behavior.
                  - "strip_id": Input items are grouped by their strip_id to create multiple composites.

    Returns:
        dict: A dictionary representing the composite tool configuration.

    Raises:
        planet.exceptions.ClientError: If group_by is not one of "order" or "strip_id".
    """
    if group_by and group_by not in ["order", "strip_id"]:
        raise ClientError(
            f"Invalid group_by value: {group_by}. Must be 'order' or 'strip_id'."
        )

    parameters = {}
    if group_by:
        parameters['group_by'] = group_by

    return _tool('composite', parameters)


def coregister_tool(anchor_item: str) -> dict:
    """Create the API spec representation of a coregister tool.

    Parameters:
        anchor_item: The item_id of the item to which all other items should be
            coregistered.
    """
    return _tool('coregister', {'anchor_item': anchor_item})


def file_format_tool(file_format: str) -> dict:
    """Create the API spec representation of a file format tool.

    Parameters:
        file_format: The format of the tool output. Either 'COG' or 'PL_NITF'.

    Raises:
        planet.specs.SpecificationException: If file_format is not one of
            'COG' or 'PL_NITF'
    """
    file_format = specs.validate_file_format(file_format)
    return _tool('file_format', {'format': file_format})


def reproject_tool(projection: str,
                   resolution: Optional[float] = None,
                   kernel: Optional[str] = None) -> dict:
    """Create the API spec representation of a reproject tool.

    Parameters:
        projection: A coordinate system in the form EPSG:n. (ex. EPSG:4326 for
            WGS84, EPSG:32611 for UTM 11 North (WGS84), or EPSG:3857 for Web
            Mercator).
        resolution: The pixel width and height in the output file. The API
            default is the resolution of the input item. This value will be in
            meters unless the coordinate system is geographic (like EPSG:4326),
            then it will be a pixel size in decimal degrees.
        kernel: The resampling kernel used. The API default is "near". This
            parameter also supports "bilinear", "cubic", "cubicspline",
            "lanczos", "average" and "mode".
    """
    parameters = dict((k, v) for k, v in locals().items() if v)
    return _tool('reproject', parameters)


def tile_tool(tile_size: int,
              origin_x: Optional[float] = None,
              origin_y: Optional[float] = None,
              pixel_size: Optional[float] = None,
              name_template: Optional[str] = None,
              conformal_x_scaling: Optional[bool] = None) -> dict:
    """Create the API spec representation of a reproject tool.

    Parameters:
        tile_size: Height and width of output tiles in pixels and lines
            (always square).
        origin_x: Tiling system x origin in projected coordinates. The API
            default is zero.
        origin_y: Tiling system y origin in projected coordinates. The API
            default is zero.
        pixel_size: Tiling system pixel size in projected coordinates. The API
            default is the pixel_size of input raster.
        name_template: A naming template for creating output tile filenames.
            The API default is "{tilex}_{tiley}.tif" resulting in filenames
            like 128_200.tif. The {tilex} and {tiley} parameters can be of the
            form {tilex:06d} to produce a fixed width field with leading zeros.
    """
    parameters = dict((k, v) for k, v in locals().items() if v)
    return _tool('tile', parameters)


def toar_tool(scale_factor: Optional[int] = None) -> dict:
    """Create the API spec representation of a TOAR tool.

    Parameters:
        scale_factor: Scale factor applied to convert 0.0 to 1.0 reflectance
            floating point values to a value that fits in 16bit integer pixels.
            The API default is 10000. Values over 65535 could result in high
            reflectances not fitting in 16bit integers.
    """
    parameters = {}
    if scale_factor:
        parameters['scale_factor'] = scale_factor
    return _tool('toar', parameters)


def harmonize_tool(target_sensor: str) -> dict:
    """Create the API spec representation of a harmonize tool.

    Parameters:
        target_sensor: A value indicating to what sensor the input asset types
            should be calibrated.

    Raises:
        planet.exceptions.ClientError: If target_sensor is not valid.
    """

    try:
        target_sensor = specs.get_match(target_sensor,
                                        specs.HARMONIZE_TOOL_TARGET_SENSORS,
                                        'target_sensor')
    except specs.SpecificationException as e:
        raise ClientError(e)

    return _tool('harmonize', {'target_sensor': target_sensor})


def band_math_tool(b1: str,
                   b2: Optional[str] = None,
                   b3: Optional[str] = None,
                   b4: Optional[str] = None,
                   b5: Optional[str] = None,
                   b6: Optional[str] = None,
                   b7: Optional[str] = None,
                   b8: Optional[str] = None,
                   b9: Optional[str] = None,
                   b10: Optional[str] = None,
                   b11: Optional[str] = None,
                   b12: Optional[str] = None,
                   b13: Optional[str] = None,
                   b14: Optional[str] = None,
                   b15: Optional[str] = None,
                   pixel_type: str = specs.BAND_MATH_PIXEL_TYPE_DEFAULT):
    """Specify an Orders API band math tool.

    The parameters of the bandmath tool define how each output band in the
    derivative product should be produced, referencing the product inputs’
    original bands. Band math expressions may not reference neighboring pixels,
    as non-local operations are not supported. The tool can calculate up to 15
    bands for an item. Input band parameters may not be skipped. For example,
    if the b4 parameter is provided, then b1, b2, and b3 parameters are also
    required.

    For each band expression, the bandmath tool supports normal arithmetic
    operations and simple math operators offered in the Python numpy package.
    (For a list of supported mathematical functions, see
    [Bandmath supported numpy math routines](https://developers.planet.com/apis/orders/bandmath-numpy-routines/)).

    One bandmath imagery output file is produced for each product bundle, with
    output bands derived from the band math expressions. nodata pixels are
    processed with the band math equation. These files have “_bandmath”
    appended to their file names.

    The tool passes through UDM, RPC, and XML files, and does not update values
    in these files.

    Parameters:
        b1-b15: An expression defining how the output band should be computed.
        pixel_type: A value indicating what the output pixel type should be.

    Raises:
        planet.exceptions.ClientError: If pixel_type is not valid.
    """  # noqa
    try:
        pixel_type = specs.get_match(pixel_type,
                                     specs.BAND_MATH_PIXEL_TYPE,
                                     'pixel_type')
    except specs.SpecificationException as e:
        raise ClientError(e)

    # e.g. {"b1": "b1", "b2":"arctan(b1)"} if b1 and b2 are specified
    parameters = dict((k, v) for k, v in locals().items() if v)
    return _tool('bandmath', parameters)


def sentinel_hub(collection_id: Optional[str] = None,
                 create_configuration: Optional[bool] = False) -> dict:
    """Specify a Sentinel Hub hosting destination."""
    params: Dict[str, Union[str, bool]] = {}
    if collection_id:
        params['collection_id'] = collection_id
    if create_configuration:
        params['create_configuration'] = create_configuration
    return {'sentinel_hub': params}



# Copyright 2020 Planet Labs, Inc.
# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from http import HTTPStatus
import httpx
import json
import os
from pathlib import Path
import respx

import pytest

from planet.auth import _SecretFile

_here = Path(os.path.abspath(os.path.dirname(__file__)))
_test_data_path = _here / 'data'


@pytest.fixture(autouse=True, scope='module')
def test_secretfile_read():
    """Returns valid auth results as if reading a secret file"""

    def mockreturn(self):
        return {'key': 'testkey'}

    # monkeypatch fixture is not available above a function scope
    # usage: https://docs.pytest.org/en/6.2.x/reference.html#pytest.MonkeyPatch
    with pytest.MonkeyPatch.context() as mp:
        mp.setattr(_SecretFile, 'read', mockreturn)
        yield


@pytest.fixture
def open_test_img():
    img_path = _test_data_path / 'test_sm.tif'
    with open(img_path, 'rb') as img:
        yield img


@pytest.fixture
def get_test_file_json():

    def func(filename):
        file_path = _test_data_path / filename
        with open(file_path, 'r') as f:
            res = json.load(f)
        return res

    return func


@pytest.fixture
def search_result(get_test_file_json):
    filename = 'search_result_20200130_093253_ssc6_u0001.json'
    return get_test_file_json(filename)


@pytest.fixture
def order_description(get_test_file_json):
    filename = 'order_description_b0cb3448-0a74-11eb-92a1-a3d779bb08e0.json'
    return get_test_file_json(filename)


@pytest.fixture
def order_request(get_test_file_json):
    filename = 'order_details_psorthotile_analytic.json'
    return get_test_file_json(filename)


@pytest.fixture
def orders_page(get_test_file_json):
    filename = 'orders_page.json'
    return get_test_file_json(filename)


@pytest.fixture
def oid():
    return 'b0cb3448-0a74-11eb-92a1-a3d779bb08e0'


@pytest.fixture
def search_id():
    return '20200130_093253_ssc6_u0001'


@pytest.fixture
def write_to_tmp_json_file(tmp_path):

    def write(data, filename):
        cc = tmp_path / filename
        with open(cc, 'w') as fp:
            json.dump(data, fp)
        return cc

    return write


@pytest.fixture
def geom_geojson():
    # these need to be tuples, not list, or they will be changed
    # by shapely
    return {
        "type": "Polygon",
        "coordinates": [
            [
                [37.791595458984375, 14.84923123791421],
                [37.90214538574219, 14.84923123791421],
                [37.90214538574219, 14.945448293647944],
                [37.791595458984375, 14.945448293647944],
                [37.791595458984375, 14.84923123791421],
            ]
        ],
    }  # yapf: disable


@pytest.fixture
def str_geom_reference():
    return "pl:features/my/water-fields-RqB0NZ5/rmQEGqm"


@pytest.fixture
def geom_reference():
    return {
        "type": "ref",
        "content": "pl:features/my/water-fields-RqB0NZ5/rmQEGqm",
    }  # yapf: disable


@pytest.fixture
def feature_geojson(geom_geojson):
    return {"type": "Feature", "properties": {}, "geometry": geom_geojson}


@pytest.fixture
def featurecollection_geojson(feature_geojson):
    return {"type": "FeatureCollection", "features": [feature_geojson]}


@pytest.fixture
def point_geom_geojson():
    return {
        "type": "Point",
        "coordinates": [37.791595458984375, 14.84923123791421]
    }


@pytest.fixture
def multipolygon_geom_geojson():
    return {
        "type":
        "MultiPolygon",
        "coordinates":
        [[[[37.791595458984375, 14.84923123791421],
          [37.90214538574219, 14.84923123791421],
          [37.90214538574219, 14.945448293647944],
          [37.791595458984375, 14.945448293647944],
          [37.791595458984375, 14.84923123791421]]]]
    }  # yapf: disable


@pytest.fixture
def anyio_backend():
    return 'asyncio'


@pytest.fixture
def mock_bundles(autouse=True, scope="session"):
    # The following bundles are not real, only used for tests
    resp = {
        "bundles": {
            "analytic_udm2": {
                "assets": {
                    "PSScene": ["ortho_analytic_4b"]
                }
            },
            "analytic_3b_udm2": {
                "assets": {
                    "PSScene": []
                }
            },
            "analytic_8b_udm2": {
                "assets": {
                    "PSScene": []
                }
            },
            "analytic_sr": {
                "assets": {
                    "SkySatScene": [], "PSScene": [], "SkySatCollect": []
                }
            },
            "analytic": {
                "assets": {
                    "SkySatScene": []
                }
            },
            "visual": {
                "assets": {
                    "PSScene": ["basic_udm2"]
                }
            }
        }
    }
    spec_url = "https://api.planet.com/compute/ops/bundles/spec"
    respx.get(spec_url).return_value = httpx.Response(HTTPStatus.OK, json=resp)



# Copyright 2022 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Test Session Configuration.

This is a script for identifying the best session configuration for fast,
reliable, and polite communication with the Planet services.
"""
import argparse
import asyncio
from collections import Counter
import inspect
import logging
import sys

import pytest

import planet
from planet import data_filter

LOGGER = logging.getLogger(__name__)


@pytest.mark.asyncio
@pytest.mark.parametrize('execution_number', range(3))
@pytest.mark.parametrize('rate_limit, max_active', [(10, 50), (10, 100),
                                                    (0, 0)])
@pytest.mark.parametrize('limit, num_concurrent', [(100, 3)])
async def test_configuration(execution_number,
                             limit,
                             num_concurrent,
                             rate_limit,
                             max_active,
                             monkeypatch):
    """Find configuration that successfully completes and runs the fastest"""
    # Setting rate_limit and max_active to zero disables limiting
    LOGGER.warning(
        'order: limit, num_concurrent, rate_limit, max_active, execution_number'
    )

    async def _search(cl):
        sfilter = data_filter.permission_filter()
        items = cl.search(['PSScene'], sfilter, limit=limit)
        return list([i async for i in items])

    await _runit(_search,
                 planet.DataClient,
                 monkeypatch,
                 num_concurrent=num_concurrent,
                 read_timeout=30,
                 rate_limit=rate_limit,
                 max_active=max_active)


@pytest.mark.asyncio
@pytest.mark.parametrize('execution_number', range(3))
@pytest.mark.parametrize('rate_limit, max_active', [(0, 0)])
@pytest.mark.parametrize('limit, num_concurrent', [(100, 3)])
async def test_api_rate_limiting(execution_number,
                                 limit,
                                 num_concurrent,
                                 rate_limit,
                                 max_active,
                                 monkeypatch):
    """Test API rate limit behavior"""
    LOGGER.warning(
        'order: limit, num_concurrent, rate_limit, max_active, execution_number'
    )

    async def _search(cl):
        sfilter = data_filter.permission_filter()
        items = cl.search(['PSScene'], sfilter, limit=limit)
        return list([i async for i in items])

    monkeypatch.setattr(planet.http, 'RETRY_EXCEPTIONS', [])
    await _runit(_search,
                 planet.DataClient,
                 monkeypatch,
                 num_concurrent=num_concurrent,
                 read_timeout=30,
                 rate_limit=rate_limit,
                 max_active=max_active)


@pytest.mark.asyncio
@pytest.mark.parametrize('execution_number', range(5))
@pytest.mark.parametrize('rate_limit, max_active', [(10, 50)])
@pytest.mark.parametrize('limit, num_concurrent', [(100, 1000)])
async def test_reliability(execution_number,
                           limit,
                           num_concurrent,
                           rate_limit,
                           max_active,
                           monkeypatch):
    """Stress test the reliability of the communication with Planet services.
    """
    LOGGER.warning(
        'order: limit, num_concurrent, rate_limit, max_active, execution_number'
    )

    async def _search(cl):
        sfilter = data_filter.permission_filter()
        items = cl.search(['PSScene'], sfilter, limit=limit)
        return list([i async for i in items])

    await _runit(_search,
                 planet.DataClient,
                 monkeypatch,
                 num_concurrent=num_concurrent,
                 read_timeout=30,
                 rate_limit=rate_limit,
                 max_active=max_active)


async def _runit(func,
                 client,
                 monkeypatch,
                 num_concurrent,
                 read_timeout,
                 rate_limit,
                 max_active,
                 retry_exceptions=None):
    """Run a method of a client with different session settings"""
    monkeypatch.setattr(planet.http, 'READ_TIMEOUT', read_timeout)
    monkeypatch.setattr(planet.http, 'RATE_LIMIT', rate_limit)
    monkeypatch.setattr(planet.http, 'MAX_ACTIVE', max_active)

    if retry_exceptions:
        monkeypatch.setattr(planet.http, 'RETRY_EXCEPTIONS', retry_exceptions)

    async with planet.Session() as sess:
        cl = client(sess)
        tasks = [asyncio.create_task(func(cl)) for _ in range(num_concurrent)]

        res = await asyncio.gather(*tasks, return_exceptions=True)
        LOGGER.warning(f'Session outcomes: {sess.outcomes}')

        exceptions = [r for r in res if isinstance(r, BaseException)]
        if len(exceptions):
            task_exceptions = Counter([type(e) for e in exceptions])
            LOGGER.warning(f'Task exceptions: {task_exceptions}')
            raise exceptions[0]


if __name__ == "__main__":
    # automatically find the tests within this module for specifying as
    # possible choices for test argument
    this_module = sys.modules[__name__]
    test_options = [
        t[0] for t in inspect.getmembers(this_module, inspect.isfunction)
        if t[0].startswith('test_')
    ]

    log_levels = ('debug', 'info', 'warning', 'error', 'critical')

    parser = argparse.ArgumentParser(
        description='Session configuration tests.')
    parser.add_argument('--log-level', default='warning', choices=log_levels)
    parser.add_argument('test',
                        choices=test_options,
                        help='specify the test to run')
    args = parser.parse_args()

    cmds = [
        '--tb=line',  # we are only interested in the types of exceptions raised
        '--log-format=%(asctime)s.%(msecs)03d %(levelname)s %(message)s',
        '--log-date-format=%Y-%m-%d %H:%M:%S',
        f'--log-cli-level={args.log_level}',
        '--durations=0',  # record and report runtimes
        __file__,
        f'-k={args.test}'
    ]
    print(f'Running: pytest {" ".join(cmds)}')
    sys.exit(pytest.main(cmds))
